/*
 * $Id $
 * 
 * $Date: 2006/12/07 09:38:06 $ $Revision: 1.23 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar;

import java.io.File;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import cern.jjar.runner.JJarRunner;
import cern.jjar.runner.ReleaseRunner;

/**
 * An ant task for publishing jars
 * 
 *  -<b>publish</b>- publish jars in the development repository. Actually
 * all files (but not directories) are copied to the dev area. Each product has
 * DEV link. <br/>
 * Parameters:
 * <ul>
 * <li><i>remoterepository</i>- URL to the development repository. It has to
 * be passed with file protocol (e.g.: file:////web5.cern.ch/acccodev/)
 * <li><i>localrepository</i>- Path to the directory with jars/zips/...
 * which are to be published
 * <li><i>localdescriptor</i>- path to the file on the local disc (e.g.: c:/projects/bic/project.xml)
 * </ul>
 * 
 * @author Lionel Mestre
 * @author Grzegorz Kruk
 * @version $Id: ReleaseTask.java,v 1.23 2006/12/07 09:38:06 enes Exp $
 */
public class ReleaseTask extends AbstractJJarTask {

  private String filedirectory;
  private String defaultdependencyversion = "PRO";
  private String numberofversionstokeep;
  private boolean resolvealias;
  private boolean keepallversions;
  private File localdescriptor;
  private String localrepository;
  
  private static final String NUMBER_OF_VERSIONS_TO_KEEP = "release.numberofversionstokeep";
  private static final String VERSIONS_TO_REMOVE = "product.versions.to.remove";
  private static final String UPDATE_NEXT_VERSION = "release.update.nextversion";
  private static final String NO_VERSIONS_TO_DELETE = "empty";

  //
  // -- CONSTRUCTORS -----------------------------------------------
  //

  public ReleaseTask() {}

  //
  // -- PUBLIC METHODS -----------------------------------------------
  //

  public final void setDefaultdependencyversion(String defaultVersion) {
    this.defaultdependencyversion = defaultVersion;
  }

  public final void setNumberofversionstokeep(String numberofversionstokeep) {
    this.numberofversionstokeep = numberofversionstokeep;
  }

  public final void setResolvealias(boolean resolvealias) {
    this.resolvealias = resolvealias;
  }

  public final void setKeepallversions(boolean keepallversions) {
    this.keepallversions = keepallversions;
  }

  public void setLocaldescriptor(File localdescriptor) {
    this.localdescriptor = localdescriptor;
  }

  /**
   * @param localrepository The localrepository to set.
   */
  public final void setLocalrepository(String localrepository) {
    this.localrepository = localrepository;
  }

  /**
   * @param filedirectory The filedirectory to set.
   */
  public final void setFiledirectory(String filedirectory) {
    this.filedirectory = filedirectory;
  }

  protected JJarRunner createRunner() {
    ReleaseRunner runner = new ReleaseRunner(this);
    runner.setNumberOfVersionsToKeep(parseNumberOfVersionsParameter(numberofversionstokeep));
    runner.setKeepAllVersions(keepallversions);
    runner.setResolvealias(resolvealias);
    runner.setLocaldescriptor(localdescriptor);
    runner.setLocalrepository(localrepository);
    runner.setFiledirectory(filedirectory);
    runner.setRepositoryElements(getRepositoryElements());
    runner.setDefaultdependencyversion(defaultdependencyversion);
    return runner;
  }

  protected void processResult(Object result) {
    final Map resultMap = (Map) result;
    final List versionsToRemove = (List) resultMap.get(ReleaseRunner.KEY_VERSIONS_TO_REMOVE_LIST);          
    String concatenatedString = "";
    for (Iterator i = versionsToRemove.iterator(); i.hasNext();) {
        if (!"".equals(concatenatedString)){
          concatenatedString = concatenatedString + ",";
        }
        concatenatedString = concatenatedString + "" + (String) i.next();
    }
    if("".equals(concatenatedString)){
        concatenatedString = NO_VERSIONS_TO_DELETE;
    }
    getProject().setNewProperty(VERSIONS_TO_REMOVE, concatenatedString);
    final Boolean equalProAndNextVersion = (Boolean) resultMap.get(ReleaseRunner.KEY_UPDATE_NEXT_VERSION_FLAG);
    if (equalProAndNextVersion.booleanValue() == true){
      getProject().setNewProperty(UPDATE_NEXT_VERSION, "true");
    }
  }
  
  private int parseNumberOfVersionsParameter(String numberOfversions) {
    log("NumberOfVersionsToKeep from class is: " + this.numberofversionstokeep);
    log("NumberOfVersionsToKeep from input arg is: " + numberOfversions);
    int intValue = 0;
    try {
      intValue = Integer.parseInt(numberOfversions);
    } catch (NumberFormatException nfe) {
      throw new RuntimeException("Property " +NUMBER_OF_VERSIONS_TO_KEEP+ " is not valid ["+numberOfversions+"], should be an integer.");
    }
    return intValue;
  }
}
