/*
 * $Id $
 * 
 * $Date: 2006/12/17 12:37:59 $ $Revision: 1.12 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.repository;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/**
 * <p>
 * Simple class to figure out ordered dependency lists. Basic idea is that you
 * load it with datum consisting of a set consisting of a package name and list
 * of packages that it's dependent upon.
 * </p>
 * <p>
 * Then, you should be able to ask for the dependencies for any package placed
 * in there.
 * </p>
 * 
 * @author Geir Magnusson Jr. (original idea from Jakarta)
 * @author Lionel Mestre
 * @version $Id: DependencyEngine.java,v 1.12 2006/12/17 12:37:59 enes Exp $
 */
public class DependencyEngine {

  private static final int INITIAL_DEPTH = 1;

  /**
   * A comparator that can be used to compare two sub contexts on their start time
   */
  private static final Comparator DEPENDENCIES_COMPARATOR = new Comparator() {
    public final int compare(Object object1, Object object2) {
      Dependency d1 = (Dependency) object1;
      Dependency d2 = (Dependency) object2;
      return d1.getProduct().compareTo(d2.getProduct());
    }
  };

  private Repository repository;

  public DependencyEngine(Repository repository) {
    this.repository = repository;
  }

  /**
   * returns a list of dependencies for a list of projects, each project being
   * included in the dependencies list
   * 
   * @param dependenciesToResolve packages to get dependency list for
   * @param excludeTarget boolean to control exclusion of each target package
   * @param defaultVersion the default version to use if no version specified
   * @return List list of dependencies, in order
   */
  public Dependency[] getDependencies(Dependency[] dependenciesToResolve, boolean excludeTarget,
      Version defaultVersion, boolean forceAll) {
    Map dependencies = new HashMap();
    Map visitedMap = new HashMap();

    for (int i = 0; i < dependenciesToResolve.length; i++) {
      if (dependenciesToResolve[i].getVersionAsObject() == null) {
        dependenciesToResolve[i].setVersion(defaultVersion);
      }
      resolveDependencies(dependenciesToResolve[i], dependencies, visitedMap, INITIAL_DEPTH, dependenciesToResolve[i]
          .isLocal(), dependenciesToResolve[i].isExcluded(), forceAll);
    }

    ArrayList result = new ArrayList();
    Iterator iterator = dependencies.values().iterator();
    while (iterator.hasNext()) {
      AnnotatedDependency element = (AnnotatedDependency) iterator.next();
      displayVersionList(element);
      result.add(element.getDependency());
    }
    if (excludeTarget) {
      for (int i = 0; i < dependenciesToResolve.length; i++) {
        result.remove(dependenciesToResolve[i]);
      }
    }
    Collections.sort(result, DEPENDENCIES_COMPARATOR);
    return (Dependency[]) result.toArray(new Dependency[result.size()]);
  }

  private void displayVersionList(AnnotatedDependency annotatedDependency) {
    Version[] versionList = annotatedDependency.getVersionList();
    if (versionList.length != 0) {
      StringBuffer sb = new StringBuffer();
      sb.append("!!! Version conflict for ");
      formatString(sb, annotatedDependency.getDependency().getProduct(), 30);
      sb.append(" kept: ");
      formatString(sb, annotatedDependency.getDependency().getVersionAsObject().toString(), 9);
      sb.append(" other(s): ");
      for (int i = 0; i < versionList.length; i++) {
        if (i != 0) {
          sb.append(", ");
        }
        sb.append(versionList[i]);
      }
     // System.out.println(sb.toString());
    }
  }

  private static void formatString(StringBuffer sb, String string, int length) {
    sb.append(string);
    int n = Math.max(0, length - string.length());
    for (int i = 0; i < n; i++) {
      sb.append(' ');
    }
  }

  private static void addDependencyToMap(Map annotatedDependencies, Dependency dependency, int depth, boolean local,
      boolean excluded) {
    String projectName = dependency.getProduct();
    AnnotatedDependency annotatedDependency = (AnnotatedDependency) annotatedDependencies.get(projectName);
    if (annotatedDependency != null) {
      annotatedDependency.completeDependency(dependency, depth, local, excluded);
    } else {
      annotatedDependencies.put(projectName, new AnnotatedDependency(dependency, depth, local, excluded));
    }
  }

  private void resolveDependencies(Dependency dependencyToResolve, Map dependencies, Map visitedMap, int depth,
      boolean local, boolean excluded, boolean forceAll) {
    //System.out.println("dependencyToResolve = "+dependencyToResolve+"  depth="+depth+"  local="+local+"  excluded="+excluded);
    String productName = dependencyToResolve.getProduct();
    Version productVersion = dependencyToResolve.getVersionAsObject();
    Product product = repository.getProduct(productName, productVersion);
    if (product == null) {
      System.out.println("!!! -> product [" + productName + ", " + productVersion
          + "] is not defined in any repository");
      return;
    }

    // we take as a version the one that we found
    productVersion = product.getVersion();
    dependencyToResolve.setVersion(productVersion);
    if (visitedMap.containsKey(productName)) {
      Version visitedVersion = (Version) visitedMap.get(productName);

      if (visitedVersion.equals(productVersion)) {
        //Product already added..
        //System.out.println("!!! -> A cycle has been detected with " + productName + ":" + v);
        return;
      }
      //Huston, we have a problem...
      System.out.println("!!! -> the dependencies contain the same product with two different versions : newest will be taken.");
      System.out.println("        - " + productName + " [" + visitedVersion + ", " + productVersion + "]");
      if (visitedVersion.compareTo(productVersion) > 0)
        return;
    }
    // the dependency is local if it is dependent on a local dependency or if it is marked as local
    local |= dependencyToResolve.isLocal();
    excluded |= dependencyToResolve.isExcluded();
    addDependencyToMap(dependencies, dependencyToResolve, depth, local, excluded);
    visitedMap.put(productName, productVersion);
    Dependency[] dependenciesToFollow = product.getDependenciesAsArray();
    for (int i = 0; i < dependenciesToFollow.length; i++) {
      // we don't follow local dependencies beyond depth 1
      if (forceAll || !dependenciesToFollow[i].isLocal()) {
        resolveDependencies(dependenciesToFollow[i], dependencies, visitedMap, depth + 1, local, excluded, forceAll);
      }
    }
    visitedMap.remove(productName);
  }

  public static class AnnotatedDependency {

    private Dependency dependency;
    private int depth;
    private Set versionList;

    public AnnotatedDependency(Dependency dependency, int depth, boolean local, boolean excluded) {
      this.dependency = (Dependency) dependency.clone();
      this.dependency.setExcluded(excluded);
      this.dependency.setLocal(local);
      this.depth = depth;
      this.versionList = new HashSet();
    }

    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString() {
      return dependency.toString() + "[" + depth + "]\n";
    }

    public Dependency getDependency() {
      return dependency;
    }

    public int getDepth() {
      return depth;
    }

    public Version[] getVersionList() {
      return (Version[]) versionList.toArray(new Version[versionList.size()]);
    }

    public void completeDependency(Dependency newDependency, int depth, boolean local, boolean excluded) {
      Version currentVersion = getDependency().getVersionAsObject();
      Version newVersion = newDependency.getVersionAsObject();
      //System.out.println("depth="+depth+"  currentVersion = "+currentVersion+"  newVersion="+newVersion);
      int n = currentVersion.compareTo(newVersion);
      if (this.depth == INITIAL_DEPTH) {
        // the directive at the direct level of dependencies are the strongest
        if (n != 0) {
          versionList.add(newVersion);
        }
      } else if (depth == INITIAL_DEPTH) {
        // the directive at the direct level of dependencies are the strongest
        if (n != 0) {
          this.dependency.setVersion(newVersion);
          versionList.add(currentVersion);
          versionList.remove(newVersion);
        }
        this.dependency.setExcluded(excluded);
        this.dependency.setLocal(local);
      } else {
        // if it is true it might get false. If it is false it will stay so
        this.dependency.setExcluded(excluded && this.dependency.isExcluded());
        this.dependency.setLocal(local && this.dependency.isLocal());
        if (n == -1) {
          // current is older than new
          this.dependency.setVersion(newVersion);
          versionList.add(currentVersion);
        } else if (n == 1) {
          // current is newer than new
          versionList.add(newVersion);
        }
      }
      if (depth < this.depth) {
        this.depth = depth;
      }
    }

  }
}
