/*
 * $Id $
 * 
 * $Date: 2006/11/02 09:06:53 $ $Revision: 1.16 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.repository;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.ArrayList;
import java.text.DateFormat;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Digester creates map of objects ProductVersionsContainer. Keys are product
 * names, values these objects. Each ProductVersionsContainer may contain many
 * versions of the same product.
 * 
 * @version $Id: JJarParser.java,v 1.16 2006/11/02 09:06:53 enes Exp $
 * @author Lionel Mestre
 * @author enes
 */
public class JJarParser {
	
	private static final DateFormat DATE_FORMAT = DateFormat.getDateInstance(DateFormat.FULL);

	private Map productVersionMap = new HashMap();
	private String prefixDirectory;
	private Document document;

	//##################  Private methods  ################################
  
	private void readDocument(Document document) throws IOException {
	    Element root = document.getDocumentElement();
	    
	    prefixDirectory = readAttribute(root, "directory");
	    NodeList elements = root.getElementsByTagName("product");
	    int n = elements.getLength();
	    for (int i = 0; i < n; i++) {
	    	Product product = readProduct((Element) elements.item(i));
	    	addProduct(product);
	    }
	  }

	private Product readProduct(Element element) {
		Product product = new Product();
	    product.setName(readAttribute(element, "name"));
	    product.setVersion(Version.valueOf(readAttribute(element, "version")));
	    product.setDirectory(readAttribute(element, "directory"));
	    String alias = readAttribute(element, "link");
	    product.setAliases(readAliases(alias));
	    Element hrefElement = readChildElement(element, "href");
	    if (hrefElement != null) {
	    	product.setHref(readNodeValue(hrefElement));
	    }
	    Element descElement = readChildElement(element, "desc");
	    if (descElement != null) {
	    	product.setDescription(readNodeValue(descElement));
	    }
	    Element shortdescElement = readChildElement(element, "shortdesc");
	    if (shortdescElement != null) {
	    	product.setShortDescription(readNodeValue(shortdescElement));
	    }
	    Element jarElement = readChildElement(element, "jar");
	    if (jarElement != null) {
	    	product.setJarName(readNodeValue(jarElement));
	    }
	    Element depElement = readChildElement(element, "dependencies");
	    if (depElement != null) {
	    	NodeList elements = depElement.getElementsByTagName("dep");
	    	int n = elements.getLength();
	    	for (int i = 0; i < n; i++) {
	    		product.addDependency(readDependency((Element) elements.item(i)));
	    	}
	    }
	    return product;
	}

	private ArrayList readAliases(String alias){
		ArrayList aliases = new ArrayList();
		if(alias != null){
			String[] multAlias = alias.split(",");
			for(int i=0;i<multAlias.length; i++){
				if(!"".equals(multAlias[i]))
					aliases.add(Version.valueOf(multAlias[i]));
			}
		}
		return aliases;
	}
	  
	private Dependency readDependency(Element element) {
		Dependency dep = new Dependency();
	    dep.setProduct(readAttribute(element, "product"));
	    dep.setLocal(readBooleanAttribute(element, "local"));
	    dep.setExcluded(readBooleanAttribute(element, "excluded"));
	    dep.setVersion(Version.valueOf(readAttribute(element, "version")));
	    return dep;
	}

	private Document parse(InputStream inputStream) throws IOException {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
	    factory.setValidating(false);
	    DocumentBuilder builder;
	    try {
	    	builder = factory.newDocumentBuilder();
	      
	    	return builder.parse(inputStream);
	    } catch (Exception e) {
	    	throw new IOException(e.toString());
	    }
	}

	/*
	 * Add product object.
	 */
	private void addProduct(Product product) throws IOException {
		//Not the best place for this - but the easiest
		if (product.getName() == null || product.getVersion() == null || product.getDirectory() == null) {
			throw new IOException("Not all attributes of product: " + product + " have been defined!");
		}
	    //Get container for given productName.
	    MultiVersionsProduct container = (MultiVersionsProduct) productVersionMap.get(product.getName());
	    if (container == null) {
	    	container = new MultiVersionsProduct(product.getName());
	    	productVersionMap.put(product.getName(), container);
	    }
	    //Add new product
	    container.addProduct(product);
	  }
	
	private static void printMap(Map m) {
		System.out.println("Map Size: " + m.size());
		for (java.util.Iterator i = m.keySet().iterator(); i.hasNext();) {
			String pn = (String) i.next();
			MultiVersionsProduct pvs = (MultiVersionsProduct) m.get(pn);
			System.out.println(pvs.getProductName());
			Product[] products = pvs.getProducts();
			for (int j = 0; j < products.length; j++) {
				Product p = products[j];
				System.out.println("   version(" + p.getVersion() + ") link(" + p.getAliasesAsString() + ") href("
						+ p.getHref() + ")");
				Dependency[] d = p.getDependenciesAsArray();
				for (int k = 0; k < d.length; k++) {
					System.out.println("   --> dep product(" + d[k].getProduct() + ") version(" + d[k].getVersionAsObject()
							+ ") local(" + d[k].isLocal() + ") excluded(" + d[k].isExcluded() + ")");
				}
			}
		}
	}

	private static Element readChildElement(Element element, String name) {
		NodeList elements = element.getElementsByTagName(name);
		if (elements.getLength() == 0) {
			return null;
		}
		return (Element) elements.item(0);
	}

	public static String readNodeValue(Element element) {
		Node child = element.getFirstChild();
		if (child == null)
			return null;
		return child.getNodeValue();
	}
	
	private static String readAttribute(Element element, String attributeName) {
		String value = element.getAttribute(attributeName);
		
		if (value == null) {
			return null;
		}
		if (value.length() == 0) {
			return null;
		}
		return value;
	}
	
	private static boolean readBooleanAttribute(Element element, String attributeName) {
		String value = readAttribute(element, attributeName);
		return "true".equals(value) || "1".equals(value) || "yes".equals(value);
	}	

	//######################### Public methods ######################################

	public JJarParser(URL url) throws IOException {
		this(url.openStream());
	}

	public JJarParser(File f) throws IOException {
		this(new FileInputStream(f));
	}

	public JJarParser(InputStream inputStream) throws IOException {
		document = parse(inputStream);
		readDocument(document);
	}
  
	public Document getDocument(){
		return document;
	}
  
	public Map getProductVersionMap() {
		return productVersionMap;
	}

	public String getPrefixDirectory() {
		return prefixDirectory;
	}

	public static void main(String[] args) {
		try {
    	
			//URL f = new URL("file://C:/repositoryXML/pro/repository.xml");
			JJarParser parser = new JJarParser(new FileInputStream("C:\\repositoryXML\\pro\\new_repository.xml"));
			Map m = parser.getProductVersionMap();
			printMap(m);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
}