/*
 * $Id $
 * 
 * $Date: 2006/11/30 10:42:48 $ $Revision: 1.11 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.repository;

/**
 * simple class to handle JJar version information
 * 
 * should be of format <major>.<minor>-<modifier>
 * 
 * GKruk: I added linked versions: PRO, NEXT, PREV, DEV. So the version can be
 * a number as above or one of the four links
 * 
 * @author Geir Magnusson Jr. (original idea from Jakarta)
 * @author Lionel Mestre
 * @version $Id: Version.java,v 1.11 2006/11/30 10:42:48 enes Exp $
 */
public class Version implements Comparable, Cloneable {

  private static final String PRO = "PRO";
  private static final String NEXT = "NEXT";
  private static final String PREV = "PREVIOUS";
  private static final String DEV = "DEV";

  public static final Version PRODUCTION_VERSION = new Version(PRO);
  public static final Version NEXT_VERSION = new Version(NEXT);
  public static final Version PREVIOUS_VERSION = new Version(PREV);
  public static final Version DEV_VERSION = new Version(DEV);

  private String stringValue = PRO;
  private int major = -1;
  private int minor = -1;
  private int tiny = -1;
  private String modifier;

  //
  // -- CONSTRUCTORS -----------------------------------------------
  //

  private Version() {}

  private Version(String version) {
    this.stringValue = version;
    if (!isAlias()) {
      parse(version);
    }
  }

  /*
   * (non-Javadoc)
   * 
   * @see java.lang.Object#clone()
   */
  public Object clone() {
    try {
      return super.clone();
    } catch (CloneNotSupportedException e) {
      return this;
    }
  }

  private void refreshStringValue() {
    StringBuffer sb = new StringBuffer();
    boolean appended = false;
    if (major > -1) {
      sb.append(major);
      appended = true;
    }
    if (minor > -1) {
      if (appended) {
        sb.append('.');
      }
      sb.append(minor);
      appended = true;
    }
    if (tiny > -1) {
      if (appended) {
        sb.append('.');
      }
      sb.append(tiny);
      appended = true;
    }
    if (modifier != null) {
      if (appended) {
        sb.append('-');
      }
      sb.append(modifier);
    }
    stringValue = sb.toString();
  }

  public static Version valueOf(String version) {
    if (version == null) {
      return null;
    }
    if (PRO.equalsIgnoreCase(version) || "".equals(version)) {
      return PRODUCTION_VERSION;
    } else if (PREV.equalsIgnoreCase(version)) {
      return PREVIOUS_VERSION;
    } else if (NEXT.equalsIgnoreCase(version)) {
      return NEXT_VERSION;
    } else if (DEV.equalsIgnoreCase(version)) {
      return DEV_VERSION;
    } else {
      return new Version(version);
    }
  }

  public static void nextMajor(Version version) {
    if (version.modifier != null) {
      version.modifier = null;
    } else {
      version.major++;
      version.minor = 0;
      version.tiny = 0;
    }
    version.refreshStringValue();
  }

  public static void nextMinor(Version version) {
    if (version.modifier != null) {
      version.modifier = null;
    } else {
      version.minor++;
      version.tiny = 0;
    }
    version.refreshStringValue();
  }

  public static void nextTiny(Version version) {
    if (version.modifier != null) {
      version.modifier = null;
    } else {
      version.tiny++;
    }
    version.refreshStringValue();
  }

  public static void nextModifier(Version version) {
    String modifier = version.modifier;
    if (modifier == null || modifier.length() == 0) {
      return;
    }
    int n = modifier.length();
    int pos = n;
    for (int i = n - 1; i >= 0; i--) {
      if (Character.isDigit(modifier.charAt(i))) {
        pos--;
      } else {
        break;
      }
    }
    if (pos < n) {
      String firstPart = modifier.substring(0, pos);
      String secondPart = modifier.substring(pos);
      int value = Integer.parseInt(secondPart);
      value++;
      version.modifier = firstPart + value;
    } else {
      version.modifier = version.modifier + "0";
    }
    version.refreshStringValue();
  }

  //
  // -- PUBLIC METHODS -----------------------------------------------
  //

  public boolean isAlias() {
    return stringValue.equals(PRO) || stringValue.equals(NEXT) || stringValue.equals(PREV) || stringValue.equals(DEV);
  }

  public int hashCode() {
    if (stringValue != null) {
      return stringValue.hashCode();
    }
    return super.hashCode();
  }

  public boolean equals(Object obj) {
    if (!obj.getClass().equals(getClass())) {
      return false;
    }
    Version v = (Version) obj;
    if (stringValue == null && v.stringValue == null)
      return true;
    return stringValue != null ? stringValue.equals(v.stringValue) : v.stringValue.equals(stringValue);
  }

  public String stringValue() {
    return stringValue;
  }

  public String toString() {
    return stringValue;
  }

  public int compareTo(Object other) {
    Version version = (Version) other;
    if (equals(version)) {
      return 0;
    }
    if (isAlias() && version.isAlias()) {
      if (this.equals(DEV_VERSION)) {
        return 1;
      }
      if (this.equals(NEXT_VERSION)) {
        if (version.equals(DEV_VERSION)) {
          return 1;
        }
        return -1;
      }
      if (this.equals(PRODUCTION_VERSION)) {
        if (version.equals(DEV_VERSION) && version.equals(PRODUCTION_VERSION)) {
          return 1;
        }
        return -1;
      }
      if (this.equals(PREVIOUS_VERSION)) {
        return -1;
      }
    }
    if (isAlias() || version.isAlias()) {
      throw new RuntimeException("Logical Version is not comparable to numeric version !");
    }

    if (major < version.major) {
      return -1;
    }
    if (major > version.major) {
      return 1;
    }
    // 2 majors ==
    if (minor < version.minor) {
      return -1;
    }
    if (minor > version.minor) {
      return 1;
    }
    // 2 minors ==
    if (tiny < version.tiny) {
      return -1;
    }
    if (tiny > version.tiny) {
      return 1;
    }
    // 2 tinys ==
    if (modifier != null && version.modifier == null) {
      return -1;
    }
    if (modifier == null && version.modifier != null) {
      return 1;
    }
    return 0;
  }

  //
  // -- PRIVATE METHODS -----------------------------------------------
  //

  private void parse(String version) {
	  //\p{Punct}: Punctuation: One of !"#$%&'()*+,-./:;<=>?@[\]^_`{|}~
    String[] items = version.split("\\p{Punct}");
    if (items.length == 0) {
      return;
    }
    if (isModifier(items[0])) {
      modifier = items[0];
    } else {
      major = extractInt(items[0]);
    }
    if (items.length == 1) {
      return;
    }
    if (isModifier(items[1])) {
      modifier = items[1];
    } else {
      minor = extractInt(items[1]);
    }
    if (items.length == 2) {
      return;
    }
    if (isModifier(items[2])) {
      modifier = items[2];
    } else {
      tiny = extractInt(items[2]);
    }
    if (items.length == 3) {
      return;
    }
    if (isModifier(items[3])) {
      modifier = items[3];
    } else {
      //tiny = items[3];
    }
  }

  private int extractInt(String fragment) {
    try {
      return Integer.parseInt(fragment);
    } catch (NumberFormatException e) {
      // find the first non digit char
      for (int i = 0; i < fragment.length(); i++) {
        if (!Character.isDigit(fragment.charAt(i))) {
          modifier = fragment.substring(i);
          return Integer.parseInt(fragment.substring(0, i));
        }
      }
      return 0;
    }
  }

  private boolean isModifier(String s) {
    if (s.length() == 0) {
      return false;
    }
    char c = s.charAt(0);
    return !Character.isDigit(c);
  }

  private int getMajor() {
    return major;
  }

  private int getMinor() {
    return minor;
  }

  private int getTiny() {
    return tiny;
  }

  private String getModifier() {
    return modifier;
  }

  public static void main(String args[]) {
    Version jv = new Version("9M4");
    System.out.println("jv: " + jv);
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    nextModifier(jv);
    System.out.println("nextModifier = " + jv);
    jv = new Version("2.1-rc1");
    System.out.println("jv: " + jv);
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    nextModifier(jv);
    System.out.println("nextModifier = " + jv);
    jv = new Version("3");
    System.out.println("jv: " + jv);
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    nextModifier(jv);
    System.out.println("nextModifier = " + jv);
    jv = new Version("4.2.5");
    System.out.println("jv: " + jv);
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    nextModifier(jv);
    System.out.println("nextModifier = " + jv);
    jv = new Version("8.5.6-dev");
    System.out.println("jv: " + jv);
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    nextModifier(jv);
    System.out.println("nextModifier = " + jv);
    jv = new Version("98.5654.62465-beta");
    System.out.println("jv: " + jv);
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    nextModifier(jv);
    System.out.println("nextModifier = " + jv);
    jv = new Version("beta");
    System.out.println("jv: " + jv);
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    nextModifier(jv);
    System.out.println("nextModifier = " + jv);
    jv = new Version("1.0");
    System.out.println(jv + " compareTo 1.1 is " + jv.compareTo(new Version("1.1")));
    System.out.println(jv + " compareTo 0.9.5 is " + jv.compareTo(new Version("0.9.5")));
    System.out.println(jv + " compareTo 1.0 is " + jv.compareTo(new Version("1.0")));
    System.out.println(jv + " compareTo 1.0.0 is " + jv.compareTo(new Version("1.0.0")));
    System.out.println(jv + " compareTo 1.0-beta is " + jv.compareTo(new Version("1.0-beta")));
    
    jv = new Version("1.2.0");
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    System.out.println(jv + " compareTo 1.1.0 is " + jv.compareTo(new Version("1.1.0")));
    System.out.println(jv + " compareTo 1.11.0 is " + jv.compareTo(new Version("1.11.0")));
    System.out.println(jv + " compareTo 1.2.1 is " + jv.compareTo(new Version("1.2.1")));
    
    jv= new Version("1.11.0.0.4");
    System.out.println(jv.getMajor() + " : " + jv.getMinor() + " : " + jv.getTiny() + " : " + jv.getModifier());
    
  }

}
