/*
 * $Id $
 * 
 * $Date: 2006/05/11 12:53:02 $ $Revision: 1.8 $ $Author: lmestre $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.runner;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;

import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import cern.jjar.repository.LocalDescriptorXML;
import cern.jjar.repository.Product;
import cern.jjar.repository.Version;
import cern.jjar.resources.Resources;
import cern.jjar.utils.Logger;

/**
 * Allow to increase or to set the version number in the descriptor file.
 * Uses increase to increase the version number (selecting MAJOR, MINOR, TINY or MODIFIER).
 * Uses version to set the version to a given value.
 * 
 * @author Lionel Mestre
 */
public class VersionUpdateRunner extends AbstractJJarRunner {
  
  public static final String MAJOR = "major";
  public static final String MINOR = "minor";
  public static final String TINY = "tiny";
  public static final String MODIFIER = "modifier";

  /* Local descriptor file */
  private File localdescriptor;

  private String increase;
  private String version;

  //
  // -- CONSTRUCTORS -----------------------------------------------
  //

  /* Creates a new instance of PublishRunner */
  public VersionUpdateRunner(Logger logger) {
    super(logger);
  }

  //
  // -- Setters -----------------------------------------------
  //

  /**
   * @param localdescriptor The localdescriptor to set.
   */
  public final void setLocaldescriptor(File localdescriptor) {
    this.localdescriptor = localdescriptor;
  }

  /**
   * @param increase The increase to set.
   */
  public final void setIncrease(String increase) {
    this.increase = increase;
  }
  
  /**
   * @param version The version to set.
   */
  public final void setVersion(String version) {
    this.version = version;
  }
  
  //
  // -- PRIVATE METHODS -----------------------------------------------
  //

  public Object run() throws Exception {
    checkParameters();
    LocalDescriptorXML localDescriptorXML = new LocalDescriptorXML(localdescriptor);
    Product[] products = localDescriptorXML.getProducts();
    if (products.length == 0) {
      return null;
    }
    Version currentVersion = products[0].getVersion();
    String previousVersion = currentVersion.stringValue();
    if (version != null) {
      currentVersion = Version.valueOf(version);
    } else if (MAJOR.equalsIgnoreCase(increase)) {
      Version.nextMajor(currentVersion);
    } else if (MINOR.equalsIgnoreCase(increase)) {
      Version.nextMinor(currentVersion);
    } else if (TINY.equalsIgnoreCase(increase)) {
      Version.nextTiny(currentVersion);
    } else if (MODIFIER.equalsIgnoreCase(increase)) {
      Version.nextModifier(currentVersion);
    }
    System.out.println("Changing version from "+previousVersion+" to "+currentVersion);
    updateVersion(currentVersion.stringValue());
    return null;
  }

  private void updateVersion(String version) throws TransformerException, IOException {
    Templates templates = Resources.getUpdateVersionTemplate();
    Transformer t = templates.newTransformer();
    t.setParameter("version", version);
    File output = new File(localdescriptor.getParentFile(), "_" + localdescriptor.getName());
    File localRepositoryBackup = new File(localdescriptor.getParentFile(), localdescriptor.getName() + ".bak");
    OutputStream outputStream = new FileOutputStream(output); 
    StreamResult outStream = new StreamResult(outputStream);
    t.transform(new StreamSource(localdescriptor), outStream);
    outputStream.close();
    localRepositoryBackup.delete();
    localdescriptor.renameTo(localRepositoryBackup);
    output.renameTo(localdescriptor);
  }
  
  public static void main(String[] args) throws TransformerException, IOException {
    VersionUpdateRunner runner = new VersionUpdateRunner(null);
    runner.setLocaldescriptor(new File("C:\\lionel\\cvs\\cmmnbuild\\cern-jjar\\product.xml"));
    runner.updateVersion("4.0.0");
  }

  private void checkParameters() throws Exception {
    if (localdescriptor == null)
      throw new Exception("localdescriptor has to be set!");
    if ((increase == null && version == null) || (increase != null && version != null))
      throw new Exception("Either version or increase has to be set!");
  }

}