/* * $Id $
 * 
 * $Date: 2005/03/23 09:40:20 $ $Revision: 1.5 $ $Author: lmestre $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar;

import java.io.File;

import org.apache.tools.ant.types.Path;

import cern.jjar.runner.GetRunner;
import cern.jjar.runner.JJarRunner;

/**
 * An ant task for getting jars (downloading jars to the local repository).
 * There are three possible repositories: production repository, development
 * repository and third-party repository. The path to each of these has to be
 * always passed as a URL e.g.: production repository:
 * http://www.cern.ch/xxx/~pcrops/production development repository:
 * file:////web5.cern.ch/test-release/dev 3-rd party:
 * http://www.cern.ch/xxx/~pcrops/thirdparty
 * 
 * It is important, that if repository file is to be changed, the path has to
 * be passed with "file" protocol, if it is only read it can be either "http"
 * or "file". It will be specified in the description of each command.
 * 
 * <br/><br/>Commands description: <br/>-<B>getJars</B>-
 * <ul>
 * <li><i>localrepository</i>- Path to the local repository dir where the
 * jars are to be downloaded (e.g.: c:/projects/bic/lib)
 * <li><i>localdescriptor</i>- path to the file on the local disc (e.g.:
 * c:/projects/bic/project.xml)
 * <li><i>repository</i>- URL of the production repository. It can be
 * either local or remote (e.g.: http://slwww.cern.ch/xxx/production,
 * file:////web3.cern.ch/production)
 * <li><i>devrepository</i>- URL of the development repository, it also can
 * be either local or remote (http or file protocol).
 * <li><i>thirdpartyrepository</i>- URL or the third-party repository. As
 * above. <br/><br/>Optional parameters: <br/>
 * <li><i>onlydependencies</i>- if true - the project jars are not
 * dowloaded (only dependent products). The default value is <i>true</i>
 * <li><i>keepversion</i>- if true - the jars are dowloaded without the
 * version number. The default value is false.
 * <li><i>localprojectpriority</i>- if true - JJar creates in the
 * localrepository dir a directory /localproject with a file whose name <br/>
 * points the path on the local disc with dependency project. The path starts
 * from the root directory of the cvsroot e.g.:
 * japc@japc-ext@japc-ext-cmw.project <br/>The default value is false.
 * <li><i>product</i>- Name of the product for which dependences jars
 * should be fetched. This property should be passed if there is no local
 * descriptor file (localdescriptor property is not set).
 * <li><i>version</i>- Version of product mentioned above. The default
 * value is PRO (production version).
 * <li><i>existingversion</i>- Version residing on the local disc. existing
 * version (if defined) is compared with <i>version</i>- if is lower that
 * version to fetch, the jars are not downloaded.
 * <li><i>defaultdependencyversion</i>- name of default version, when
 * ommited in dependency definition.
 * </ul>
 * 
 * @author Lionel Mestre
 * @author Grzegorz Kruk
 * @version $Id: GetTask.java,v 1.5 2005/03/23 09:40:20 lmestre Exp $
 */
public class GetTask extends AbstractJJarTask {

  private String pathrefid;

  /* Local descriptor file */
  private File localdescriptor;

  /* Name of the product */
  private String product;

  /* Version of the product */
  private String version;

  /* Keep version while fetching jars */
  private boolean keepversion;

  /**
   * Create file pointing the path to dependent product stored on the local
   * machine.
   */
  private boolean localprojectpriority;

  /* Local repository dir. Always will be lib, except publishing */
  private String localrepository = "lib";

  /* Indicates if product should be fetched together with its dependencies */
  private boolean onlydependencies = true;

  /* Version of the product existing in the local repository */
  private String existingversion;

  /**/
  private String defaultdependencyversion = "PRO";

  /**/
  private boolean createlink;

  //
  // -- CONSTRUCTORS -----------------------------------------------
  //

  public GetTask() {}

  //
  // -- PUBLIC METHODS -----------------------------------------------
  //

  public void setPathrefid(String pathrefid) {
    this.pathrefid = pathrefid;
  }

  public void setLocaldescriptor(File localdescriptor) {
    this.localdescriptor = localdescriptor;
  }

  public void setProduct(String product) {
    this.product = product;
  }

  public void setVersion(String version) {
    this.version = version;
  }

  public void setLocalrepository(String localrepository) {
    this.localrepository = localrepository;
  }

  public void setKeepversion(boolean keepversion) {
    this.keepversion = keepversion;
  }

  public void setLocalprojectpriority(boolean localprojectpriority) {
    this.localprojectpriority = localprojectpriority;
  }

  public void setOnlydependencies(boolean onlydependencies) {
    this.onlydependencies = onlydependencies;
  }

  public void setCreatelink(boolean createlink) {
    this.createlink = createlink;
  }

  public void setExistingversion(String existingversion) {
    this.existingversion = existingversion;
  }

  public void setDefaultdependencyversion(String defaultVersion) {
    this.defaultdependencyversion = defaultVersion;
  }

  protected JJarRunner createRunner() {
    GetRunner runner = new GetRunner(this);
    runner.setExistingversion(existingversion);
    runner.setKeepversion(keepversion);
    runner.setLocaldescriptor(localdescriptor);
    runner.setLocalprojectpriority(localprojectpriority);
    runner.setLocalrepository(localrepository);
    runner.setOnlydependencies(onlydependencies);
    runner.setProduct(product);
    runner.setVersion(version);
    runner.setDefaultdependencyversion(defaultdependencyversion);
    runner.setRepositoryElements(getRepositoryElements());
    runner.setCreatelink(createlink);
    return runner;
  }

  protected void processResult(Object result) {
    String[] addedPaths = (String[]) result;
    if (pathrefid != null && addedPaths.length > 0) {
      // create an antpath
      Path antpath = new Path(getProject());
      for (int i = 0; i < addedPaths.length; i++) {
        antpath.createPath().setPath(addedPaths[i]);
      }
      getProject().addReference(pathrefid, antpath);
    }
  }

}
