/*
 * $Id $
 * 
 * $Date: 2006/11/30 10:42:48 $ $Revision: 1.17 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.repository;

import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.w3c.dom.Document;

import cern.jjar.JJarRuntimeException;
import cern.jjar.runner.CleanRunner;

/**
 * A Repository holds one or several Stores.
 * 
 * @author Grzegorz Kruk
 */
public class RepositoryXML implements Repository {

    private final Version defaultVersion;

    private List repositoryStores = new ArrayList();

    private DependencyEngine depEngine = null;

    public RepositoryXML(String defaultVersion) {
        depEngine = new DependencyEngine(this);
        if (defaultVersion == null) {
            this.defaultVersion = Version.PRODUCTION_VERSION;
        } else {
            this.defaultVersion = Version.valueOf(defaultVersion);
        }
    }

    public RepositoryXML() {
        this(null);
    }

    public void addStore(RepositoryStore repositoryXMLStore) {
        repositoryStores.add(repositoryXMLStore);
    }

    /*
     * (non-Javadoc)
     * 
     * @see cern.jjar.repository.Repository#getProductionStore()
     */
    public RepositoryStore getProductionStore() {
        RepositoryStore store = null;
        for (int i = 0; i < repositoryStores.size(); i++) {
            if (((RepositoryStore) repositoryStores.get(i)).isProduction()) {
                if (store != null) {
                    throw new JJarRuntimeException(
                            "There are multiple production repository stores.");
                }
                store = (RepositoryStore) repositoryStores.get(i);
            }
        }
        if (store == null) {
            throw new JJarRuntimeException(
                    "Could not find production repository store.");
        }
        return store;
    }

    public Dependency[] getDependencies(Dependency dependencyToResolve,
            boolean excludeTarget, Version defaultVersion) {
        return depEngine.getDependencies(
                new Dependency[] { dependencyToResolve }, excludeTarget,
                defaultVersion, false);
    }

    public Dependency[] getDependencies(Dependency[] dependenciesToResolve,
            boolean excludeTargets, Version defaultVersion) {
        return depEngine.getDependencies(dependenciesToResolve, excludeTargets,
                defaultVersion, false);
    }

    /**
     * Return URL of product's jar or null if product does not exist or product
     * does not contain jar.
     */
    public URL getProductFileUrl(String productName, Version version) {
        ProductStore productStore = getProductStoreAnyVersion(productName,
                version);
        if (productStore != null) {
            return productStore.store.getProductFileUrl(productStore.product);
        }
        // the product does not exist
        return null;
    }

    /**
     * Return URL of product's jar or null if product does not exist or product
     * does not contain jar.
     */
    public String getProductFilePath(String productName, Version version) {
        ProductStore productStore = getProductStoreAnyVersion(productName,
                version);
        if (productStore != null) {
            return productStore.store.getProductFilePath(productStore.product);
        }
        // the product does not exist
        return null;
    }

    /**
     * Returns Product object with given name and version or null if not found
     * in the repository.
     */
    public Product getProduct(String productName, Version version) {
        // System.out.println("--> productName="+productName+"
        // version="+version);
        ProductStore productStore = getProductStoreAnyVersion(productName,
                version);
        // System.out.println("--> productStore ="+productStore);
        if (productStore != null) {
            return productStore.product;
        }
        return null;
    }

    public URL getProductionDescriptorRootDirURL() {
        return getProductionStore().getStoreRootDirURL();
    }

    public Document getDocument() {
        return getProductionStore().getDocument();
    }
    
    public void printOut(Object[] list){
        System.out.println("--------------------------------");
        for (int i = 0; i < list.length; i++) {
            System.out.println(list[i].toString());
        }
        System.out.println("--------------------------------");
    }
    
    public void printOut(Collection list){
        System.out.println("--------------------------------");
        for (Iterator iter = list.iterator(); iter.hasNext();) {
            Product product = (Product) iter.next();
            System.out.println(product);
        }
        System.out.println("--------------------------------");
    }

    public Product[] computeObsoleteVersionsForProduct(Product product) {
        // assume product is always defined in the production store
        final MultiVersionsProduct mvp = getProductionStore().getProduct(product.getName());
        // in case of an initial release of a new product 
        if (mvp == null) {
            return new Product[0];
        }
        
        printOut(mvp.getProducts());
        
        //removes all versions which has an alias. These are not obsolete
        Product[] versionsFilteredOnAlias = filterVersionsOnAlias(mvp.getProducts());
        
        printOut(versionsFilteredOnAlias);
        
        Set allDependecies = getAllDependenciesInAllRepositories();
        //versions which are not represented in allDependencies are considered obsolete
        List obsoleteVersions = filterVersionsOnDependencies(versionsFilteredOnAlias, allDependecies);
        printOut(obsoleteVersions);
        return (Product[]) obsoleteVersions.toArray(new Product[obsoleteVersions.size()]);
    }

    private List filterVersionsOnDependencies(Product[] versionsToFilter, Set allDependecies) {
        List filteredList = new ArrayList();
        for (int i = 0; i < versionsToFilter.length; i++) {
            String productName = versionsToFilter[i].getName();
            Version productVersion = versionsToFilter[i].getVersion();
            
            //to find if the version of the product we are checking is osbsolete, we create a new instance 
            //of Dependency based on the productinfo and check if it is in the list
            if (!allDependecies.contains(new Dependency(productName,
                    productVersion))) {
                filteredList.add(versionsToFilter[i]);
            }
        }
        return filteredList;
    }

    private Product[] filterVersionsOnAlias(Product[] products) {
        List list = new ArrayList();

        for (int i = 0; i < products.length; i++) {
            if (!products[i].hasAlias()) {
                list.add(products[i]);
            }
        }
        return (Product[]) list.toArray(new Product[list.size()]);
    }

    private Set getAllDependenciesInAllRepositories() {
        final Set allDeps = new HashSet();
        for (int i = 0; i < repositoryStores.size(); i++) {
            RepositoryStore store = (RepositoryStore) repositoryStores.get(i);
            MultiVersionsProduct[] allMultiVersionedProducts = store
                    .getProducts();

            for (int j = 0; j < allMultiVersionedProducts.length; j++) {
                Product[] multiVersionedProduct = allMultiVersionedProducts[j]
                        .getProducts();

                for (int k = 0; k < multiVersionedProduct.length; k++) {
                    Dependency[] deps = multiVersionedProduct[k]
                            .getDependenciesAsArray();
                    allDeps.addAll(Arrays.asList(deps));
                }
            }
        }
        return allDeps;
    }

    public Product[] computeObsoleteProducts(Product product) {
        Map nonObsoleteProducts = new HashMap();
        for (int i = 0; i < repositoryStores.size(); i++) {
            RepositoryStore store = (RepositoryStore) repositoryStores.get(i);
            MultiVersionsProduct[] products = store.getProducts();
            computeNonObsoleteProductsImpl(nonObsoleteProducts, products);
        }
        Map obsoleteProducts = new HashMap();
        for (int i = 0; i < repositoryStores.size(); i++) {
            RepositoryStore store = (RepositoryStore) repositoryStores.get(i);
            MultiVersionsProduct[] products = store.getProducts();
            computeObsoleteProductsImpl(product, obsoleteProducts, products,
                    nonObsoleteProducts);
        }
        return (Product[]) obsoleteProducts.values().toArray(
                new Product[obsoleteProducts.size()]);
    }

    private void computeObsoleteProductsImpl(Product productToClean,
            Map obsoleteProducts, MultiVersionsProduct[] mvProducts,
            Map nonObsoleteProducts) {
        for (int i = 0; i < mvProducts.length; i++) {
            MultiVersionsProduct mvProduct = mvProducts[i];
            Product[] products = mvProduct.getProducts();
            for (int j = 0; j < products.length; j++) {
                Product product = products[j];

                if (!containsProduct(nonObsoleteProducts, product)
                        && (productToClean.getName()).equals(products[j]
                                .getName())) {
                    fillProductMap(obsoleteProducts, product);
                }
            }
        }
    }

    private void computeNonObsoleteProductsImpl(Map nonObsoleteProducts,
            MultiVersionsProduct[] mvProducts) {
        for (int i = 0; i < mvProducts.length; i++) {
            MultiVersionsProduct mvProduct = mvProducts[i];
            Product[] products = mvProduct.getProducts();
            for (int j = 0; j < products.length; j++) {
                Product product = products[j];
                if (product.hasAlias()
                        && !containsProduct(nonObsoleteProducts, product)) {
                    // if (product.getName().equals("laser-core")) {
                    // System.out.println(product);
                    // }
                    addNonObsoleteProduct(nonObsoleteProducts, product);
                }
            }
        }
    }

    private void addNonObsoleteProduct(Map nonObsoleteProducts, Product product) {
        String productName = product.getName();
        Version version = product.getVersion();
        Dependency targetDependency = new Dependency(productName, version);
        Dependency[] dependencies = depEngine.getDependencies(
                new Dependency[] { targetDependency }, true, defaultVersion,
                true);
        fillProductMap(nonObsoleteProducts, product);
        for (int i = 0; i < dependencies.length; i++) {
            Dependency dependency = dependencies[i];
            Product depProduct = getProduct(dependency.getProduct(), dependency
                    .getVersionAsObject());
            if (depProduct != null) {
                fillProductMap(nonObsoleteProducts, depProduct);
            }
        }
    }

    private static void fillProductMap(Map products, Product product) {
        String productName = product.getName();
        Version version = product.getVersion();
        String key = productName + "_" + version.stringValue();
        products.put(key, product);
    }

    private static boolean containsProduct(Map products, Product product) {
        String productName = product.getName();
        Version version = product.getVersion();
        String key = productName + "_" + version.stringValue();
        return products.get(key) != null;
    }

    /*
     * Returns Product object with given name and version or null if not found
     * in the repository.
     */
    private ProductStore getProductStoreAnyVersion(String productName,
            Version version) {
        if (version == null) {
            version = defaultVersion;
        }
        ProductStore productStore = null;
        ProductStore devProductStore = null;
        do {
            if (version.equals(Version.DEV_VERSION)) {
                devProductStore = getProductStoreForOneVersion(productName,
                        version);
            } else {
                productStore = getProductStoreForOneVersion(productName,
                        version);
            }
            version = downgradeVersion(version);
        } while (productStore == null && version != null);
        if (devProductStore != null) {
            if (productStore == null) {
                return devProductStore;
            }
            Version devVersion = devProductStore.product.getVersion();
            Version nonDevVersion = productStore.product.getVersion();
            if (nonDevVersion.compareTo(devVersion) == 1) {
                // the version not dev is more recent than the dev one
                return productStore;
            }
            return devProductStore;
        }
        return productStore;
    }

    private Version downgradeVersion(Version version) {
        if (version.equals(Version.DEV_VERSION)) {
            return Version.NEXT_VERSION;
        }
        if (version.equals(Version.NEXT_VERSION)) {
            return Version.PRODUCTION_VERSION;
        }
        if (version.equals(Version.PRODUCTION_VERSION)) {
            return Version.PREVIOUS_VERSION;
        }
        return null;
    }

    /*
     * Returns Product object with given name and version or null if not found
     * in the repository.
     */
    private ProductStore getProductStoreForOneVersion(String productName,
            Version version) {
        Product p = null;
        // Look for product version
        for (Iterator i = repositoryStores.iterator(); i.hasNext();) {
            RepositoryStore store = (RepositoryStore) i.next();
            p = store.getProduct(productName, version);
            if (p != null) {
                return new ProductStore(store, p);
            }
        }
        // Product not found
        return null;
    }

    private class ProductStore {
        Product product;

        RepositoryStore store;

        ProductStore(RepositoryStore store, Product product) {
            this.store = store;
            this.product = product;
        }
    }

    public static void main(String[] args) {
        try {
            RepositoryXML r = new RepositoryXML(); // new
            // URL("http://test-netbeans.webtest.cern.ch/test-netbeans/test/Release"));
            // r.addStore("http://test-netbeans.webtest.cern.ch/test-netbeans/test/Release/pcrops/",
            // "http://slwww.cern.ch/~pcrops/apps/production/pcrops/");

            System.out.println("Product openIDE: "
                    + r.getProduct("openide", Version.valueOf("3.5")));
            System.out.println("ProductNotExisting: "
                    + r.getProduct("abcde", Version.PRODUCTION_VERSION));

            Dependency d = new Dependency();
            d.setProduct("bic-ui");
            d.setVersion(Version.DEV_VERSION);
            Dependency[] deps = r.getDependencies(d, false,
                    Version.PRODUCTION_VERSION);
            System.out.println("Dependencies for bic-japc: ");
            for (int i = 0; i < deps.length; i++) {
                System.out.println(" "
                        + deps[i]
                        + " urlPath: "
                        + r.getProductFileUrl(deps[i].getProduct(), deps[i]
                                .getVersionAsObject()));
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
}
