/*
 * $Id $
 * 
 * $Date: 2006/11/30 10:42:48 $ $Revision: 1.25 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.runner;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import cern.jjar.repository.Product;
import cern.jjar.repository.Repository;
import cern.jjar.repository.Version;
import cern.jjar.utils.Logger;

/**
 * @author lmestre
 */
public class CleanRunner extends AbstractJJarRunner {

  private Document repositoryDocument;
  private Product[] productsToClean;
  private Repository repository;
  private int numberOfVersionsToKeep;

  /**
   * This Set contains all versions that might be removed for all products specified in product.xml.
   * This list must be filtered to make sure not to remove a directory which contains a product still
   * in use
   */
  private Set allVersionsToRemove = new HashSet();

  /**
   * This Set contains the versions of product(s) that will be removed
   */
  //private Set versionsToRemove = new HashSet();

  /* Creates a new instance of CleanRunner */
  public CleanRunner(Logger logger) {
    super(logger);
  }

  //
  // -- PUBLIC METHODS -----------------------------------------------
  //
  /*
   * Used in JJar; cleaning from commandline
   */
  public void setUpCleanRunner(String productName, String newRepositoryFilePath) throws Exception {

    repository = lookupRepository();
    Product product = repository.getProduct(productName, Version.valueOf(defaultdependencyversion));
    if (product == null)
      throw new RuntimeException("Product not found in repository");

    setDocumentToClean(repository.getDocument());
    setProductsToClean(new Product[]{product});
  }

  public void setRepository(Repository repository) {
    this.repository = repository;
  }

  public void setDocumentToClean(Document document) {
    repositoryDocument = document;
  }

  public void setNumberOfVersionsToKeep(int noVersions) {
    numberOfVersionsToKeep = noVersions;
  }

  public void setProductsToClean(Product[] products) {
    productsToClean = products;
  }

  public Document getCleanedDocument() {
    return repositoryDocument;
  }

  /**
   * The method must take into account that several products might be defined in the product.xml. When looking
   * for versions to remove, it is important to make sure that for a version to be removed, it must be 
   * obsolete for all the products.
   * 
   * @return directoriesToRemove A list of all the directories to remove in the repository
   */
  public Object run() throws Exception {
    checkParameters();

    Set prodSpecificVersionsToRemove = new HashSet();

    for (int i = 0; i < productsToClean.length; i++) {
      Product[] obsoleteVersions = repository.computeObsoleteVersionsForProduct(productsToClean[i]);
      Arrays.sort(obsoleteVersions, Product.PRODUCT_COMPARATOR);
      prodSpecificVersionsToRemove.add(keepVersions(obsoleteVersions));
    }
    final Set versionsToRemove = resolveVersionsToRemove(prodSpecificVersionsToRemove);
    final List dirsList = getSortedDirsList(versionsToRemove);
    if (!versionsToRemove.isEmpty()) {
      log("versionsToRemove: ");
      cleanDocument(versionsToRemove);
      log("directoriesToRemove: " + dirsList);
    }
    return dirsList;
  }

  //
  //---- private methods ---------------------------------------------------------
  //
  /**
   * This method filters the list of versions to remove, and allows only the versions that
   * are obsolete for all products in the product.xml to be deleted. 
   * 
   * @param productLists list of products
   * @return dirsToRemove The directories to remove/delete from the server
   */
  private Set resolveVersionsToRemove(Set productLists) {
    Set versionsToRemove = new HashSet();
    for (Iterator i = allVersionsToRemove.iterator(); i.hasNext();) {
      boolean isContained = false;
      Product refProduct = (Product) i.next();
      for (Iterator j = productLists.iterator(); j.hasNext();) {
        List tempList = (ArrayList) j.next();
        for (int k = 0; k < tempList.size(); k++) {
          isContained = false;
          String curVersion = ((Product) tempList.get(k)).getVersion().stringValue();
          if (curVersion.equals(refProduct.getVersion().stringValue())) {
            isContained = true;
            break;
          }
        }
        if (!isContained)
          break;
      }
      if (isContained) {
        Product productToRemove = refProduct;
        versionsToRemove.add(productToRemove);
      }
    }
    return versionsToRemove;
  }

  private List getSortedDirsList(Set productsToRemove) {
    final Set dirsList = new HashSet(productsToRemove.size());
    for (Iterator iter = productsToRemove.iterator(); iter.hasNext();) {
      final Product productToRemove = (Product) iter.next();
      dirsList.add(productToRemove.getDirectory() + "/" + productToRemove.getVersion());
    }
    // sort resulting dirs list
    final String[] dirPaths = (String[]) dirsList.toArray(new String[dirsList.size()]);
    Arrays.sort(dirPaths);
    return Arrays.asList(dirPaths);
  }

  /**
   * Removes the versionsToRemove from the repository.xml
   * 
   * @throws Exception
   */
  private void cleanDocument(Set versionsToRemove) throws Exception {
    Product[] toRemove = (Product[]) versionsToRemove.toArray(new Product[versionsToRemove.size()]);
    Element root = repositoryDocument.getDocumentElement();

    XPath xpath = XPathFactory.newInstance().newXPath();

    for (int i = 0; i < toRemove.length; i++) {
      //not in use at the moment. may be used with java 1.5
      //			XPathVariableResolver resolver = new XPathVariableHandler((Product)versionsToRemove[i]);
      //			xpath.setXPathVariableResolver(resolver);

      log(toRemove[i].toString());
      String expression = "//product[@name=\"" + toRemove[i].getName() + "\"][@version=\"" + toRemove[i].getVersion()
          + "\"]";
      Element element = findNode(xpath, root, expression);
      if (element != null) {
        root.removeChild(element);
      }
    }
  }

  private Element findNode(XPath xpath, Element root, String expression) throws Exception {
    Element element = (Element) xpath.evaluate(expression, root, XPathConstants.NODE);

    if (element == null)
      throw new Exception("Problem while cleaning the repository.xml. The node was not found");

    return element;
  }

  private void checkParameters() throws Exception {
    if (repository == null)
      throw new Exception("There is no repository directory defined!");
  }

  /**
   * Removes a specified number (numberOfVersionsToKeep) of versions from the list of obsolete products. 
   * 
   * @param products The list of obsolete products
   * @return versionsToRemove The list of obsolete products - numberOfVersionsToKeep
   */
  private ArrayList keepVersions(Product[] products) {
    ArrayList versionsToRemove = new ArrayList();

    if (products.length <= numberOfVersionsToKeep)
      return versionsToRemove;

    int k = 0;
    for (int i = products.length; i > 0; i--) {
      if (i <= (products.length - numberOfVersionsToKeep)) {
        versionsToRemove.add(products[i - 1]);
        allVersionsToRemove.add(products[i - 1]);
        k++;
      }
    }
    return versionsToRemove;
  }

}