/*
 * $Id $
 * 
 * $Date: 2006/11/15 16:29:39 $ $Revision: 1.28 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.runner;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;
import java.util.ArrayList;
import java.util.List;

import javax.swing.DefaultDesktopManager;

import cern.jjar.repository.Dependency;
import cern.jjar.repository.LocalDescriptorXML;
import cern.jjar.repository.Product;
import cern.jjar.repository.Repository;
import cern.jjar.repository.Version;
import cern.jjar.utils.JJarUtils;
import cern.jjar.utils.Logger;
import cern.jjar.utils.Transport;

/**
 * @author Lionel
 */
public class GetRunner extends AbstractJJarRunner {

  public static final String PROJECT_DEPENDENCY_DIR = "/localproduct";
  public static final String PROJECT_DEPENDENCY_EXTENSION = ".product";
  public static final char PROJECT_DIR_SEPARATOR = '@';

  private static final String EXCLUDED_DEPENDENCY_DIR = "/excluded";
  private static final String EXCLUDED_DEPENDENCY_EXTENSION = ".excluded";

  /* Local descriptor file */
  private File localdescriptor;

  /* Name of the product */
  private String product;

  /* Version of the product */
  private String version;

  /* Keep version while fetching jars */
  private boolean keepversion;

  /** if true the classpath is updated with the complete list of dependent jars but nothing is fetched */
  private boolean simulated;

  /**
   * Create file pointing the path to dependent product stored on the local
   * machine.
   */
  private boolean localprojectpriority;

  /* Local repository dir. Always will be lib, except publishing */
  private String localrepository;

  /* Indicates if product should be fetched together with its dependencies */
  private boolean onlydependencies;

  /* Version of the product existing in the local repository */
  private String existingversion;

  private boolean resolvealias;

  /**
   * This property indicates making links instead of copying files while
   * fetching jars (getJars) This property MAY be only useful while releasing -
   * instead of copying files, better is to make links. THIS FEATURE WAS NOT
   * TESTED.
   */
  private boolean createlink;

  private Repository repository;

  private Product targetProduct;

  //
  // -- CONSTRUCTORS -----------------------------------------------
  //

  public GetRunner(Logger logger) {
    super(logger);
  }

  private void checkParameters() throws Exception {
    if (getRepositoryElement() == null)
      throw new Exception("There is no repository directory defined!");
  }

  //
  // -- Setters -----------------------------------------------
  //

  /**
   * @param createlink The makelink to set.
   */
  public final void setCreatelink(boolean createlink) {
    this.createlink = createlink;
  }

  /**
   * @param existingversion The existingversion to set.
   */
  public final void setExistingversion(String existingversion) {
    this.existingversion = existingversion;
  }

  /**
   * @param keepversion The keepversion to set.
   */
  public final void setKeepversion(boolean keepversion) {
    this.keepversion = keepversion;
  }

  /**
   * @param resolvealias The resolvealias to set.
   */
  public final void setResolvealias(boolean resolvealias) {
    this.resolvealias = resolvealias;
  }

  /**
   * @param localdescriptor The localdescriptor to set.
   */
  public final void setLocaldescriptor(File localdescriptor) {
    this.localdescriptor = localdescriptor;
  }

  /**
   * @param localprojectpriority The localprojectpriority to set.
   */
  public final void setLocalprojectpriority(boolean localprojectpriority) {
    this.localprojectpriority = localprojectpriority;
  }

  /**
   * @param localrepository The localrepository to set.
   */
  public final void setLocalrepository(String localrepository) {
    this.localrepository = localrepository;
  }

  /**
   * @param onlydependencies The onlydependencies to set.
   */
  public final void setOnlydependencies(boolean onlydependencies) {
    this.onlydependencies = onlydependencies;
  }

  /**
   * @param product The product to set.
   */
  public final void setProduct(String product) {
    this.product = product;
  }

  /**
   * @param version The version to set.
   */
  public final void setVersion(String version) {
    this.version = version;
  }

  /**
   * @param simulated The simulated to set.
   */
  public final void setSimulated(boolean simulated) {
    this.simulated = simulated;
  }

  public Object run() throws Exception {
    //Init - if repository root dirs were not defined
    //They are the same as repository dirs.
    checkParameters();
    if (!simulated) {
      displaySettings();
    }

    //Load repository object
    repository = lookupRepository();

    List productsList = new ArrayList();
    Dependency[] deplist = null;
    // look if the local descriptor exist
    if (localdescriptor != null && localdescriptor.exists()) {
      // use the local descriptor
      LocalDescriptorXML localDescriptorXML = new LocalDescriptorXML(localdescriptor);
      targetProduct = localDescriptorXML.getProduct();
      /*
       * dependenciesToResolve should not take into consideration the product. if it does, if several products
       * are present in the product.xml, it will only fetch the dependencies for the first product.
       */
//      Dependency[] dependenciesToResolve = (product == null || product.length() == 0) ? localDescriptorXML
//          .getDependencies() : localDescriptorXML.getDependencies(product);
      Dependency[] dependenciesToResolve = localDescriptorXML.getDependencies();
      deplist = repository.getDependencies(dependenciesToResolve, false, Version.valueOf(defaultdependencyversion));
    } else {
      // use the remote package definition
      if (product == null) {
        throw new Exception("Need to specify a package to fetch");
      }
      if (version == null) {
        version = defaultdependencyversion;
      }
      targetProduct = repository.getProduct(product, Version.valueOf(version));
      if (targetProduct == null) {
        throw new Exception("Couldn't find the product " + product + ":" + version);
      }
      if (existingversion != null && version != null) {
        Version existing = Version.valueOf(existingversion);
        Version current = targetProduct.getVersion();
        if (existing.compareTo(current) >= 0) {
          log(product + " existing version " + existingversion + " is uptodate");
          return new String[0];
        }
      }
      deplist = repository.getDependencies(new Dependency(product, Version.valueOf(version)), onlydependencies, Version
          .valueOf(defaultdependencyversion));
    }
    // process the dependencies that we have retrieved
    if (!simulated) {
      for (int i = 0; i < deplist.length; i++) {
        processDependency(deplist[i].getProduct(), deplist[i].getVersionAsObject(), deplist[i].isLocal(), deplist[i]
            .isExcluded(), productsList);
      }
    } else {
      for (int i = 0; i < deplist.length; i++) {
        Dependency dep = deplist[i];
        String path = repository.getProductFilePath(dep.getProduct(), dep.getVersionAsObject());
        if (!dep.isExcluded() && path != null) {
          if (!keepversion) {
            path = removeVersionFromName(path);
          }
          //System.out.println(path);
          if (!resolvealias && (dep.getVersionAsObject() == null || (! dep.getVersionAsObject().isAlias()))) {
            Product p = repository.getProduct(dep.getProduct(), dep.getVersionAsObject());
            path = replaceVersionByAlias(path, p, defaultdependencyversion);
          }
          productsList.add(path);
        }
      }
    }
    String[] result = new String[productsList.size()];
    productsList.toArray(result);
    return result;
  }

  //
  // -- PRIVATE METHODS -----------------------------------------------
  //

  private void displaySettings() {
    if (localdescriptor != null && localdescriptor.exists()) {
      log("Fetching products defined in " + localdescriptor.toString());
    } else {
      log("Fetching product " + product);
    }
    log("Remote repositories: ");
    RepositoryElement[] repositoryElements = getRepositoryElements();
    for (int i = 0; i < repositoryElements.length; i++) {
      log("    --> " + repositoryElements[i]);
    }
    log("Local repository : ");
    log("    --> " + localrepository);

    log(".");
    log(" - default dependency version is " + defaultdependencyversion);

    if (keepversion) {
      log(" - keeping version in the name of the fetched jars");
    }
    if (onlydependencies) {
      log(" - dependency-only mode, not fetching target");
    } else {
      log(" - fetching target and all dependencies");
    }
    if (localprojectpriority) {
      log(" - Priority to local projects over jars");
    }
    if (simulated) {
      log(" - Simulated mode : nothing if fetched");
    }
    if (createlink) {
      log(" - Symbolic link will be created in the local repository");
    }
    log(".");
  }

  private static String formatMessage(String name, Version version, boolean local, boolean excluded, int length) {
    StringBuffer sb = new StringBuffer(length);
    sb.append("-> Looking for ");
    sb.append(name);
    if (version != null) {
      sb.append("-");
      sb.append(version.stringValue());
    }
    String qualifier = (excluded ? "X." : "..") + (local ? "local" : ".....");
    int n = sb.length();
    for (int i = n; i < length - qualifier.length(); i++) {
      sb.append(".");
    }
    sb.append(qualifier);
    sb.append("./");
    return sb.toString();
  }

  private void processDependency(String dependencyName, Version dependencyVersion, boolean isLocal, boolean isExcluded,
      List productsList) throws IOException {
    String msg = formatMessage(dependencyName, dependencyVersion, isLocal, isExcluded, LOG_LINE_LENGTH);
    URL dependenceURL = repository.getProductFileUrl(dependencyName, dependencyVersion);
    if (dependenceURL == null) {
      log(msg + " WARNING The product does not exist");
      return;
    }

    if (localprojectpriority) {
      createLocalProjectPriorityFile(dependencyName);
    }

    boolean depRepositoryLocal = "file".equals(dependenceURL.getProtocol());
    File repDependenceFile = null;
    URLConnection repDependenceURLConn = null;
    long depRepositoryLastModified = 0;

    if (depRepositoryLocal) {
      repDependenceFile = new File(dependenceURL.getPath());
      if (!repDependenceFile.exists()) {
        log(msg + " WARNING couldn't find file " + repDependenceFile);
        return;
      }
      depRepositoryLastModified = repDependenceFile.lastModified();
    } else {
      repDependenceURLConn = dependenceURL.openConnection();
      depRepositoryLastModified = repDependenceURLConn.getLastModified();
      if (depRepositoryLastModified == 0) {
        log(msg + "WARNING couldn't find file " + dependenceURL);
        return;
      }
    }

    String localFileNameWithVersion = getFileName(dependenceURL);
    String localFileName = localFileNameWithVersion;
    if (!keepversion) {
      localFileName = removeVersionFromName(localFileName);
    }

    File depLocalFile = new File(localrepository + "/" + localFileName);
    long depLocalLastModified = depLocalFile.lastModified();

    if (depRepositoryLastModified > depLocalLastModified) {
      log(msg + "GET>" + localFileNameWithVersion);
      //
      if (depRepositoryLocal) {
        if (createlink) {
          //This is only for Release purpose - instead of copying required jars
          // - it is better just to make links
          Transport.makeLink(repDependenceFile, localrepository + "/" + localFileName);
        } else {
          Transport.fetchJar(repDependenceFile, localrepository + "/" + localFileName);
        }
      } else {
        Transport.fetchJar(repDependenceURLConn, localrepository + "/" + localFileName);
      }

      if (isExcluded) {
        createExcludedDependency(localFileName);
      }
      productsList.add(localrepository + "/" + localFileName);
    } else {
      log(msg + "skip");
    }
  }

  private void createExcludedDependency(String dependenceLocalPathname) throws IOException {
    File destDirectory = new File(localrepository + EXCLUDED_DEPENDENCY_DIR);
    File file = new File(destDirectory, dependenceLocalPathname + EXCLUDED_DEPENDENCY_EXTENSION);
    //log(" Package " + dependenceLocalPathname + " is excluded, adding a local
    // entry");
    destDirectory.mkdirs();
    file.createNewFile();
  }

  private void createLocalProjectPriorityFile(String dependencyProductName) throws IOException {
    String targetProductDir = targetProduct.getDirectory();
    //If the path is created using \ - replace it into /
    targetProductDir = targetProductDir.replace('\\', '/');

    String[] dirs = targetProductDir.split("/");
    StringBuffer relativePath = new StringBuffer("/../");
    for (int i = 0; i < dirs.length; i++) {
      relativePath.append("../");
    }

    File rootDir = new File(localrepository + relativePath);

    Product localDependencyProduct = repository.getProduct(dependencyProductName, Version.DEV_VERSION);
    if (localDependencyProduct == null) {
      //The dependent product is not defined
      //It is impossible to determine its directory
      return;
    }

    File localDependencyProductDir = new File(localDependencyProduct.getDirectory());
    File dependencyDir = new File(rootDir, localDependencyProductDir.toString());

    //If the dependency product directory does not exist - exit
    if (!dependencyDir.exists())
      return;

    String localDependencyFileName = localDependencyProductDir.toString().replace(File.separatorChar,
        PROJECT_DIR_SEPARATOR);
    File destDirectory = new File(localrepository + PROJECT_DEPENDENCY_DIR);
    File file = new File(destDirectory, localDependencyFileName + PROJECT_DEPENDENCY_EXTENSION);
    destDirectory.mkdirs();
    file.createNewFile();
  }

  private static String replaceVersionByAlias(String path, Product p, String defaultdependencyversion) {
    if (p == null || !p.hasAlias() || p.hasAlias(Version.PREVIOUS_VERSION)) {
      return path;
    }
    
    /*
     * If a product has more than one alias, it will first check if it has the alias corresponding to the defaultdependencyversion.
     * If not, it will return the other. If no alias is found, it will return null.
    */
    Version alias = p.getAlias(defaultdependencyversion);
    String dir = JJarUtils.concatenatePath(p.getDirectory(), "");
    int pos = path.indexOf(dir);
    if (pos < 0) {
      return path;
    }
    int startVersionPosition = pos + dir.length();
    int endVersionPosition = path.indexOf('/', startVersionPosition);
    if (endVersionPosition < 0) {
      return path;
    }
    String newpath = path.substring(0, startVersionPosition) + alias.stringValue() + path.substring(endVersionPosition);
    //System.out.println("--> " + newpath);
    return newpath;
  }

  private static String removeVersionFromName(String name) {
    int extensionDotIndex = name.lastIndexOf(".");
    int slashIndex = name.lastIndexOf("/");
    if (extensionDotIndex == -1) {
      return name;
    }
    String extension = name.substring(extensionDotIndex);
    int lastDashIndex = name.lastIndexOf("-");
    if (lastDashIndex == -1 || lastDashIndex < slashIndex) {
      return name;
    }
    // the last dash is either the separator of the version, the separator of
    // the version modifier
    // such as dev, beta, or the separator in the name is there is no version.
    int beforeLastDashIndex = name.lastIndexOf("-", lastDashIndex - 1);
    if (beforeLastDashIndex == -1) {
      // if there isn't another dash we have to check if the last dash was
      // separator of version or name
      // If it is version the first character after the last dash must be a
      // number and it should be a dot
      // somewhere after
      if (Character.isDigit(name.charAt(lastDashIndex + 1))) {
        return name.substring(0, lastDashIndex) + extension;
      }
      return name;
    }
    // if there is another dash, is is either the version separator or a
    // separator in the name
    if (Character.isDigit(name.charAt(beforeLastDashIndex + 1))) {
      return name.substring(0, beforeLastDashIndex) + extension;
    }
    // if the other dash is not version separator we back up to the last one
    if (Character.isDigit(name.charAt(lastDashIndex + 1))) {
      return name.substring(0, lastDashIndex) + extension;
    }
    return name;
  }

  //
  // -- INNER CLASSES -----------------------------------------------
  //
  /*
   * public interface Logger { public void log(String msg); }
   */
  public static void main(String[] args) {
    // TODO use a config file
    String repositoryRoot = "file://C:\\repositoryXML\\";
    try {
      GetRunner runner = new GetRunner(null);
      runner.setRepositoryElements(new RepositoryElement[]{new RepositoryElement(repositoryRoot + "pro"),
          new RepositoryElement(repositoryRoot + "__3rd_PARTY__")});//, new RepositoryElement(repositoryRoot + "release")});
      String curDir = System.getProperty("user.dir");
      File productFile = new File("C:\\repositoryXML/product.xml");
      if (!productFile.exists() || !productFile.canRead()) {
        System.err.println("ERROR: You need to have a file product.xml in the current working directory");
        System.exit(-1);
      }
      runner.setLocaldescriptor(productFile);
      runner.setSimulated(true);
      runner.setProduct("testTwo");
      runner.setLocalrepository("C:\\repositoryXML\\pro\\lib");
      runner.setDefaultdependencyversion("NEXT");
      runner.setOnlydependencies(true);
      runner.setKeepversion(true);
      runner.setLocalprojectpriority(false);
      runner.setResolvealias(false);
      //runner.setCommand(COMMAND_GET_JARS);

      String[] jars = (String[]) runner.run();
      for (int i = 0; i < jars.length; i++) {
        System.out.println("Jar: " + jars[i]);
      }
    } catch (Exception e) {
      System.out.println("MSG: " + e.getMessage());
      e.printStackTrace();
    }
  }
}