/*
 * $Id $
 * 
 * $Date: 2006/11/09 15:41:30 $ $Revision: 1.26 $ $Author: enes $
 * 
 * Copyright CERN, All Rights Reserved.
 */
package cern.jjar.runner;

import java.io.File;
import java.io.FileFilter;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.URL;
import java.text.DateFormat;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import cern.jjar.repository.LocalDescriptorXML;
import cern.jjar.repository.Product;
import cern.jjar.repository.Repository;
import cern.jjar.resources.Resources;
import cern.jjar.utils.Logger;
import cern.jjar.utils.Transport;


import com.sun.org.apache.xml.internal.serialize.*;

//import org.apache.xml.serialize.OutputFormat;
//import org.apache.xml.serialize.XMLSerializer;


import org.w3c.dom.Document;

/**
 * @author lmestre
 */
public class PublishRunner extends AbstractJJarRunner {

  /* Local descriptor file */
  protected File localdescriptor;

  protected String localrepository;

  private static final DateFormat DATE_FORMAT = DateFormat.getDateInstance(DateFormat.FULL);
  
  private static final String LOCK_FILE_NAME = "repository.lock";
  private static final long LOCK_WAITING_TIMEMILLIS = 2500;

  //
  // -- CONSTRUCTORS -----------------------------------------------
  //

  /* Creates a new instance of PublishRunner */
  public PublishRunner(Logger logger) {
    super(logger);
  }

  //
  // -- Setters -----------------------------------------------
  //

  /**
   * @param localdescriptor The localdescriptor to set.
   */
  public final void setLocaldescriptor(File localdescriptor) {
    this.localdescriptor = localdescriptor;
  }

  /**
   * @param localrepository The localrepository to set.
   */
  public final void setLocalrepository(String localrepository) {
    this.localrepository = localrepository;
  }

  //
  // -- PUBLIC METHODS -----------------------------------------------
  //

  public Object run() throws Exception {
    checkParameters();
    LocalDescriptorXML localDescriptorXML = new LocalDescriptorXML(localdescriptor);
    File localRepositoryDir = new File(localrepository);
    Product[] products = localDescriptorXML.getProducts();
    loadJarsNames(products, localRepositoryDir);

    RepositoryElement repositoryElement = getRepositoryElement();
    String remoteRepository = repositoryElement.getUrl();
    publishViaCopy(products, remoteRepository, localRepositoryDir);
    return null;
  }

  //
  // -- PRIVATE METHODS -----------------------------------------------
  //

  private void publishViaCopy(Product[] products, String remoteRepository, File localRepositoryDir)
      throws TransformerException, IOException {
    File repositoryDir = new File(new URL(remoteRepository).getPath());
    prepareDestDirs(products, repositoryDir);
    Set publishedFiles = new HashSet();
    for (int i = 0; i < products.length; i++) {
      Product p = products[i];
      String filename = p.getJarName();
      log("Publishing product [name=" + p.getName() + ", version=" + p.getVersion() + ", jar=" + filename + "] in "
          + p.getDirectory());
      File productDestDir = new File(repositoryDir, p.getDirectory());
      if (!publishedFiles.contains(filename)) {
        // not yet published
        copyJar(productDestDir, localRepositoryDir, filename);
        publishedFiles.add(filename);
      }
      registerProduct(Resources.getRegisterProductTemplate(), repositoryDir, localdescriptor, p, "DEV");
    }
  }

  private void checkParameters() throws Exception {
    if (getRepositoryElement() == null)
      throw new Exception("Remote repository has to be set!");
    if (localrepository == null)
      throw new Exception("Local repository has to be set!");
    if (localdescriptor == null)
      throw new Exception("Local descriptor has to be set!");
  }

  //
  // -- STATIC METHODS -----------------------------------------------
  //
  protected final static void loadJarsNames(Product[] localProducts, File localRepositoryDir) {
    //Load all regular files
    File[] jars = localRepositoryDir.listFiles(new FileFilter() {
      public boolean accept(File f) {
        return f.isFile();
      }
    });
    // sort the products from the longest name to the shortest
    Arrays.sort(localProducts, new ProductNameSorter());
    // for each product tries to find matching jars. The last jar found is the good one which is what we
    // want in the case several versions are present. We want to retain the last one.
    int filenameLength = 0;
    for (int i = 0; i < localProducts.length; i++) {
      Product p = localProducts[i];
      if (p.getJarName() != null) {
        // we don't perform matching : ignore as the jarname has been explicitely specified
        continue;
      }
      String productName = p.getName();
      for (int j = 0; j < jars.length; j++) {
        File jarFile = jars[j];
        if (jarFile != null && jarFile.getName().startsWith(productName)) {
          // we found a match. 
          // For instance productName could be cern-jjar and jarFile could be cern-jjar-4.0.jar
          // We want to deal with the case where there are more than one file 
          //     cern-jjar-3.0.jar, cern-jjar-4.0.jar  cern-jjar-bin-3.0.zip  cern-jjar-bin-4.0.zip
          // In this case we want to take the last file that matches that is not of greater length than a previous 
          // match
          if (filenameLength == 0 || jarFile.getName().length() <= filenameLength) {
            p.setJarName(jarFile.getName());
            filenameLength = jarFile.getName().length();
          }
          jars[j] = null; // taken
        }
      }
    }
  }

  private static void prepareDestDirs(Product[] localProducts, File repositoryDir) {
    for (int i = 0; i < localProducts.length; i++) {
      File productDestDir = new File(repositoryDir, localProducts[i].getDirectory());
      if (productDestDir.exists()) {
        File[] files = productDestDir.listFiles();
        for (int j = 0; j < files.length; j++) {
          files[j].delete();
        }
        //System.out.println("deleted all files in dir = " + productDestDir);
      } else {
        //System.out.println("creating dir = " + productDestDir);
        productDestDir.mkdirs();
      }
    }
  }

  private static void copyJar(File destDir, File localRepositoryDir, String jarName) throws IOException {
    if (jarName == null) {
      return;
    }
    File sourceFile = new File(localRepositoryDir, jarName);
    File destFile = new File(destDir, jarName);
    //System.out.println("source = " + sourceFile);
    //System.out.println("dest   = " + destFile);
    Transport.fetchJar(sourceFile, destFile.getAbsolutePath());
  }

  protected static final void registerProduct(Templates templates, File repositoryDir, File localdescriptor,
      Product product, String link) throws TransformerException, IOException {
    Transformer t = templates.newTransformer();
    if (link != null) {
      t.setParameter("link", link);
    }
    
    if (product.getJarName() != null) {
      t.setParameter("jar", product.getJarName());
    }
    t.setParameter("name", product.getName());
    t.setParameter("version", product.getVersion().stringValue());
    t.setParameter("localdescriptor", localdescriptor.getAbsolutePath());
    t.setParameter("releaseDate", DATE_FORMAT.format(new Date()));
    
    saveRepositoryFile(null, t, repositoryDir);
  }
  
  protected static final void saveRepositoryFile(Document document, Transformer t, File repositoryDir) throws TransformerException, IOException {
	  File lock = new File(repositoryDir, LOCK_FILE_NAME);
	  waitForLock(lock);
	  File outputFile = new File(repositoryDir, "_temp_" + Repository.REPOSITORY_FILE_NAME);
	  File repositoryFile = new File(repositoryDir, Repository.REPOSITORY_FILE_NAME);
	  File oldRepositoryFile = new File(repositoryDir, Repository.REPOSITORY_FILE_NAME+".old");
	  OutputStream outputStream = new FileOutputStream(outputFile); 
	  
	  if(document!=null){
		  OutputFormat format = new OutputFormat(document);
		  format.setEncoding("UTF-8");
		  format.setIndent(2);
		  format.setLineWidth(500);
		  format.setPreserveSpace(false);
		  
		  XMLSerializer seriOutput = new XMLSerializer(outputStream, format);
		  seriOutput.serialize(document);
	  }else if(t!=null){
		  StreamResult outStream = new StreamResult(outputStream);
		  t.transform(new StreamSource(repositoryFile), outStream);
	  }
	  
	  outputStream.close();
	  oldRepositoryFile.delete();
	  repositoryFile.renameTo(oldRepositoryFile);
	  outputFile.renameTo(repositoryFile);
	  lock.delete();
  }
  
  private static void waitForLock(File lock) throws IOException {
    if (lock.exists()) {
      try {
        Thread.sleep(LOCK_WAITING_TIMEMILLIS / 2);
      } catch (InterruptedException e) {
      }
    }
    if (lock.exists()) {
      try {
        Thread.sleep(LOCK_WAITING_TIMEMILLIS);
      } catch (InterruptedException e) {
      }
    }
    if (lock.exists()) {
      throw new IOException("Cannot register product in repository.xml, the file is locked out");
    }
    lock.createNewFile();
  }

  private static final class ProductNameSorter implements Comparator {
    public int compare(Object p1, Object p2) {
      String p1Name = ((Product) p1).getName();
      String p2Name = ((Product) p2).getName();
      return p2Name.length() - p1Name.length();
    }
  }
}