/*
 * ReleaseJarsRunner.java
 * 
 * Created on February 9, 2004, 5:51 PM
 */
package cern.jjar.runner;

import java.io.File;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import cern.jjar.JJarRuntimeException;
import cern.jjar.repository.Dependency;
import cern.jjar.repository.LocalDescriptorXML;
import cern.jjar.repository.Product;
import cern.jjar.repository.Repository;
import cern.jjar.repository.Version;
import cern.jjar.utils.Logger;

/**
 * @author lmestre
 * @author enes
 */
public class ReleaseRunner extends AbstractJJarRunner {
  
  public static final String KEY_VERSIONS_TO_REMOVE_LIST = "VERSIONS_TO_REMOVE_LIST";
  public static final String KEY_UPDATE_NEXT_VERSION_FLAG = "UPDATE_NEXT_VERSION_FLAG";

  private final DateFormat dateFormat = new SimpleDateFormat("EEEE, MMMM d, yyyy HH:mm:ss");

  private CleanRunner cleanRunner;
  private File localdescriptor;
  private String localrepository;
  private String filedirectory;
  private int numberOfVersionsToKeep;
  private URL repositoryDocumentRootDirURL;
  private boolean resolvealias;
  private boolean keepAllVersions;
  private boolean equalProAndNextVersion = false;
  
  public Logger logger;

  /* Creates a new instance of PublishRunner */
  public ReleaseRunner(Logger logger) {
    super(logger);
    this.logger = logger;
  }

  /**
   * @param localdescriptor The localdescriptor to set.
   */
  public final void setLocaldescriptor(File localdescriptor) {
    this.localdescriptor = localdescriptor;
  }

  /**
   * @param localrepository The localrepository to set.
   */
  public final void setLocalrepository(String localrepository) {
    this.localrepository = localrepository;
  }

  /**
   * @param versiondirectory The versiondirectory to set.
   */
  public final void setFiledirectory(String versiondirectory) {
    this.filedirectory = versiondirectory;
  }

  /**
   * @param number The number of obsolete versions to keep.
   */
  public final void setNumberOfVersionsToKeep(int number) {
    this.numberOfVersionsToKeep = number;
  }

  /**
   * @param keepAllVersions Boolean value whether or not to clean repository
   */
  public final void setKeepAllVersions(boolean keepAllVersions) {
    this.keepAllVersions = keepAllVersions;
  }

  /**
   * @param resolvealias Indicates whether or not to translate the alias in the dependencies
   * 					 into version-numbers.
   */
  public final void setResolvealias(boolean resolvealias) {
    this.resolvealias = resolvealias;
  }   

  /**
   * @return Returns the equalProAndNextVersion.
   */
  public boolean isEqualProAndNextVersion() {
    return this.equalProAndNextVersion;
  }

  /**
   * @param equalProAndNextVersion The equalProAndNextVersion to set.
   */
  private void setEqualProAndNextVersion(boolean equalProAndNextVersion) {
    this.equalProAndNextVersion = equalProAndNextVersion;
  }

  public Object run() throws Exception {
    LocalDescriptorXML localDescriptorXML = new LocalDescriptorXML(localdescriptor);
    File localRepositoryDir = new File(localrepository);
    Product[] newProducts = localDescriptorXML.getProducts();

    PublishRunner.loadJarsNames(newProducts, localRepositoryDir);

    System.out.println("New products: ");
    for (int i = 0; i < newProducts.length; i++) {
      System.out.println("product " + (i + 1) + ": " + newProducts[i]);
    }
    System.out.println("");

    Repository repository = lookupRepository();
    verifyVersionNumber(repository, newProducts);

    Document newElements = adjustElements(localDescriptorXML.getDocument(), newProducts, repository);
    Document updatedRepositoryDocument = updateRepositoryXML(newElements, repository.getDocument(), newProducts);

    Document processedDocument = updatedRepositoryDocument;
    List versionsToRemove = new ArrayList();
    if (!keepAllVersions) {
      versionsToRemove = (List) executeCleanRunner(repository, newProducts, updatedRepositoryDocument);
      processedDocument = cleanRunner.getCleanedDocument();
    }

    repositoryDocumentRootDirURL = repository.getProductionDescriptorRootDirURL();
    PublishRunner.saveRepositoryFile(processedDocument, null, new File(repositoryDocumentRootDirURL.getPath()));
    
    final Map result = new HashMap();
    result.put(KEY_VERSIONS_TO_REMOVE_LIST, versionsToRemove);
    result.put(KEY_UPDATE_NEXT_VERSION_FLAG, new Boolean(isEqualProAndNextVersion()));
    
    return result;
  }

  private void verifyVersionNumber(Repository repository, Product[] products) throws IllegalArgumentException {
    for (int i = 0; i < products.length; i++) {
      Product product = repository.getProduct(products[i].getName(), products[i].getVersion());
      if (product != null ){
        if((products[i].getDirectory()).equals(product.getDirectory())){
          throw new IllegalArgumentException("The versionnumber: " + products[i].getVersion() + " already exists "
            + "for this product and cannot be used twice. Please update the versionnumber");
        }
      }
    }
  }

  private Object executeCleanRunner(Repository repository, Product[] newProducts, Document updatedRepositoryDocument)
      throws Exception {
    cleanRunner = new CleanRunner(logger);
    cleanRunner.setRepository(repository);
    cleanRunner.setProductsToClean(newProducts);
    cleanRunner.setDocumentToClean(updatedRepositoryDocument);
    cleanRunner.setNumberOfVersionsToKeep(this.numberOfVersionsToKeep);

    return cleanRunner.run();
  }

  private Document adjustElements(Document document, Product[] products, Repository repository) throws Exception {
    Element root = document.getDocumentElement();
    String expression;
    Element element;

    XPath xpath = XPathFactory.newInstance().newXPath();

    for (int i = 0; i < products.length; i++) {
      //This is not in use at the moment. may be used with java 1.5
      //XPathVariableResolver resolver = new XPathVariableHandler(products[i]);
      //xpath.setXPathVariableResolver(resolver);

      expression = "//product[@name=\"" + products[i].getName() + "\"]";
      Element productElement = findNode(xpath, root, expression);
      if (productElement == null)
        continue;
      productElement.setAttribute("link", defaultdependencyversion);

      expression = "//product[@name=\"" + products[i].getName() + "\"]/jar";
      element = findNode(xpath, productElement, expression);
      if (element != null)
        productElement.removeChild(element);
      Element jar = document.createElement("jar");
      jar.appendChild(document.createTextNode(filedirectory + "/" + products[i].getJarName()));
      productElement.insertBefore(jar, productElement.getFirstChild());

      Element date = document.createElement("releaseDate");
      date.appendChild(document.createTextNode(dateFormat.format(new Date())));
      productElement.appendChild(date);

      expression = "//product[@name=\"" + products[i].getName() + "\"]/application";
      element = findNode(xpath, productElement, expression);
      if (element != null)
        productElement.removeChild(element);

      Dependency[] deps = products[i].getDependenciesAsArray();
      for (int j = 0; j < deps.length; j++) {
        //This is not in use at the moment. may be used with java 1.5
        //XPathVariableResolver depResolver = new XPathVariableHandler(deps[j]);
        //xpath.setXPathVariableResolver(depResolver);

        expression = "//product[@name=\"" + products[i].getName() + "\"]/dependencies/dep[@product=\""
            + deps[j].getProduct() + "\"]";
        //expression = "/dependencies/dep[@product=\""+deps[j].getProduct()+"\"]";
        Element depElement = findNode(xpath, productElement, expression);

        Dependency d = deps[j];
        Version v = d.getVersionAsObject() == null ? Version.valueOf(defaultdependencyversion) : d.getVersionAsObject();
        Product dp = repository.getProduct(d.getProduct(), v);

        if (dp == null)
          throw new JJarRuntimeException("Dependencyproduct: \"" + d.getProduct()
              + "\" was not found in repository. Verify the spelling of the dependency in \"product.xml\"");

        if (resolvealias || !dp.hasAlias()) {
          depElement.setAttribute("version", dp.getVersion().stringValue());
        } else {
          depElement.setAttribute("version", dp.getAlias(defaultdependencyversion).stringValue());
        }
      }
    }
    return document;
  }

  private Document updateRepositoryXML(Document newElements, Document document, Product[] products) throws Exception {
    Element root = document.getDocumentElement();

    XPath xpath = XPathFactory.newInstance().newXPath();

    for (int i = 0; i < products.length; i++) {
      //This is not in use at the moment. may be used with java 1.5
      //XPathVariableResolver resolver = new XPathVariableHandler(products[i]);
      //xpath.setXPathVariableResolver(resolver);

      if (this.defaultdependencyversion.equals("PRO")) {
        updateElementsWhenDependecyIsPRO(products[i], xpath, root);
      } else if (this.defaultdependencyversion.equals("NEXT")) {
        updateElementsWhenDependecyIsNEXT(products[i], xpath, root);
      }
    }
    //insert the new product
    String expression = "//product";
    Element element = findNode(xpath, root, expression);

    NodeList elements = newElements.getElementsByTagName("product");
    for (int i = 0; i < elements.getLength(); i++) {
      if (isEqualProAndNextVersion()) {
        ((Element) elements.item(i)).setAttribute("link", "PRO,NEXT");
      }
      root.insertBefore(document.importNode(elements.item(i), true), element);
    }
    return document;
  }

  private void updateElementsWhenDependecyIsPRO(Product product, XPath xpath, Element root) {
    String expression;

    //remove old PREVIOUS-link
    expression = "//product[@name=\"" + product.getName() + "\"][@link='PREVIOUS']";
    Element prevElement = findNode(xpath, root, expression);
    if (prevElement != null)
      prevElement.removeAttribute("link");

    //search for old PRO-link and set it to PREVIOUS
    expression = "//product[@name=\"" + product.getName() + "\"][@link='PRO']";
    Element proElement = findNode(xpath, root, expression);
    if (proElement != null)
      proElement.setAttribute("link", "PREVIOUS");

    //search for NEXT link
    expression = "//product[@name=\"" + product.getName() + "\"][@link='NEXT']";
    Element nextElement = findNode(xpath, root, expression);
    if (nextElement != null) {
      String oldNextVersion = nextElement.getAttribute("version");
      if (product.getVersion().compareTo(Version.valueOf(oldNextVersion)) >= 0) {
        setEqualProAndNextVersion(true);
        nextElement.removeAttribute("link");        
      }
    }

    //search for PRO,NEXT link of NEXT,PRO link
    String expressionNP = "//product[@name=\"" + product.getName() + "\"][@link='NEXT,PRO']";
    String expressionPN = "//product[@name=\"" + product.getName() + "\"][@link='PRO,NEXT']";
    Element npElement = findNode(xpath, root, expressionNP);
    Element pnElement = findNode(xpath, root, expressionPN);
    if (npElement != null || pnElement != null) {
      String oldNPVersion;
      Element nextNPElement;
      if (npElement != null) {
        oldNPVersion = npElement.getAttribute("version");
        nextNPElement = npElement;
      } else {
        oldNPVersion = pnElement.getAttribute("version");
        nextNPElement = pnElement;
      }
      if (product.getVersion().compareTo(Version.valueOf(oldNPVersion)) >= 0) {
        setEqualProAndNextVersion(true);
        nextNPElement.setAttribute("link", "PREVIOUS");
      }
    }
  }

  private void updateElementsWhenDependecyIsNEXT(Product product, XPath xpath, Element root) {
    //remove old link to NEXT
    String expression = "//product[@name=\"" + product.getName() + "\"][@link='NEXT']";
    Element element = findNode(xpath, root, expression);
    if (element != null)
      element.removeAttribute("link");

    //search for PRO,NEXT link of NEXT,PRO link
    String expressionNP = "//product[@name=\"" + product.getName() + "\"][@link='NEXT,PRO']";
    String expressionPN = "//product[@name=\"" + product.getName() + "\"][@link='PRO,NEXT']";
    Element npElement = findNode(xpath, root, expressionNP);
    Element pnElement = findNode(xpath, root, expressionPN);
    if (npElement != null || pnElement != null) {
      String oldNPVersion;
      Element nextNPElement;
      if (npElement != null) {
        oldNPVersion = npElement.getAttribute("version");
        nextNPElement = npElement;
      } else {
        oldNPVersion = pnElement.getAttribute("version");
        nextNPElement = pnElement;
      }
      if (product.getVersion().compareTo(Version.valueOf(oldNPVersion)) >= 0) {
        nextNPElement.setAttribute("link", "PRO");
      }
    }
  }

  private Element findNode(XPath xpath, Element root, String expression) {

    try {
      Element element = (Element) xpath.evaluate(expression, root, XPathConstants.NODE);
      return element;
    } catch (XPathExpressionException xpex) {
      xpex.printStackTrace();
      System.out.println("xpath failed");
    } catch (Exception ex) {
      ex.printStackTrace();
    }
    return null;
  }

  public static void main(String[] args) {
    try {
      ReleaseRunner runner = new ReleaseRunner(null);
      runner.setRepositoryElements(new RepositoryElement[]{new RepositoryElement("file://C:\\repositoryXML\\pro\\"),
          new RepositoryElement("file://C:\\repositoryXML\\__3rd_PARTY__\\")});
      runner.setLocaldescriptor(new File("C:\\repositoryXML\\product.xml"));
      runner.setLocalrepository("C:\\repositoryXML\\test\\build\\dist");
      runner.setDefaultdependencyversion("PRO");
      runner.setNumberOfVersionsToKeep(0);
      runner.setFiledirectory("build/dist");

      //      runner.setLocaldescriptor(new File("C:\\java\\dev\\test\\product.xml"));
      //      runner.setLocalrepository("C:\\java\\dev\\test\\build\\dist");
      //      runner.setDefaultdependencyversion("PRO");
      //      runner.setNumberOfVersionsToKeep("2");

      runner.setResolvealias(true);
      runner.setKeepAllVersions(false);

      runner.run();
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
}