/*
 * Copyright  2001-2004 The Apache Software Foundation
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */


/*
    Modified ANT 1.6.1 input task.
    Added functionality for checking if user input matches the regular expression
    specified in regexp attribute. If not, ask for input once again (just like
    the 'validargs attribute'). 
    
    It also includes a 'release-tool-specific' modification:

    1. "definedvalue" is the name of the property that keeps the value (defined,
	in release-tool, in product.xml file) to be used as and answer,
	the question and the answer are displayed on the screen    
    2. if "release.usedefaults" property is set in the project the task will not 
	ask the question (just display it and the default answer) and take a 
	defaultvalue for the answer
 
     In above cases, when no user input is expected, "+" is added before the question
     message to indicate that the question is answered automatically.
 
    3. delay, if no input specified for a certain time, Release will exit

    Look for 'add-on' in this file to see what has been changed.
    
    - Sebastian.Oglaza@cern.ch
    
*/

package cern.release.ant.taskdef;

import java.util.Timer;
import java.util.TimerTask;
import java.util.Vector;
import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.input.InputRequest;
import org.apache.tools.ant.input.MultipleChoiceInputRequest;
import org.apache.tools.ant.util.StringUtils;
import org.apache.tools.ant.input.InputHandler;

/**
 * Reads an input line from the console.
 *
 * @author <a href="mailto:usch@usch.net">Ulrich Schmidt</a>
 * @author Stefan Bodewig
 *
 * @since Ant 1.5
 *
 * @ant.task category="control"
 */
public class MyInput extends Task {
    private String validargs = null;
    private String message = "";
    private String addproperty = null;
    private String defaultvalue = null;
    private String regexp = null;
    /* add-on */
    private String definedvalue = null;
    
    public void setRegexp(String regexp) {
	this.regexp = regexp;
    }
    
    public void setDefinedvalue(String definedvalue) {
	this.definedvalue = definedvalue;
    }
    /* add-on */
    
    
    /**
     * Defines valid input parameters as comma separated strings. If set, input
     * task will reject any input not defined as accepted and requires the user
     * to reenter it. Validargs are case sensitive. If you want 'a' and 'A' to
     * be accepted you need to define both values as accepted arguments.
     *
     * @param validargs A comma separated String defining valid input args.
     */
    public void setValidargs (String validargs) {
        this.validargs = validargs;
    }

    /**
     * Defines the name of a property to be created from input. Behaviour is
     * according to property task which means that existing properties
     * cannot be overridden.
     *
     * @param addproperty Name for the property to be created from input
     */
    public void setAddproperty (String addproperty) {
        this.addproperty = addproperty;
    }

    /**
     * Sets the Message which gets displayed to the user during the build run.
     * @param message The message to be displayed.
     */
    public void setMessage (String message) {
        this.message = message;
    }

    /**
     * Defines the default value of the property to be created from input.
     * Property value will be set to default if not input is received.
     *
     * @param defaultvalue Default value for the property if no input
     * is received
     */
    public void setDefaultvalue (String defaultvalue) {
        this.defaultvalue = defaultvalue;
    }

    /**
     * Set a multiline message.
     * @param msg The message to be displayed.
     */
    public void addText(String msg) {
        message += getProject().replaceProperties(msg);
    }

    /**
     * No arg constructor.
     */
    public MyInput () {
    }

    /**
     * Actual method executed by ant.
     * @throws BuildException
     */
    public void execute () throws BuildException {
	// add-on
	if (definedvalue != null 
	    && getProject().getProperty(definedvalue) != null) {
	    
	    System.err.println("+" + message + " " + getProject().getProperty(definedvalue));
	    if (addproperty != null) 
		getProject().setNewProperty(addproperty, 
		    getProject().getProperty(definedvalue));
	
	    return;
	}
	
	
	
	// add-on
	if (getProject().getProperty("release.usedefaults") != null
	    && defaultvalue != null) {
	    
	    System.err.println("+" + message + " " +defaultvalue);
	    if (addproperty != null) 
		getProject().setNewProperty(addproperty, defaultvalue);    
	
	    return;
	}
	
	
        if (addproperty != null
            && getProject().getProperty(addproperty) != null) {
            log("skipping " + getTaskName() + " as property " + addproperty
                + " has already been set.");
            return;
        }

        InputRequest request = null;
        if (validargs != null) {
            Vector accept = StringUtils.split(validargs, ',');
            request = new MultipleChoiceInputRequest(message, accept);
        } else {
            request = new InputRequest(message);
        }
        Timer t = new Timer();
        t.schedule(new TimerTask() {
                public void run() {
                    getProject().fireBuildFinished(new BuildException("No input received, exiting.."));
                    System.exit(1);
                    
                }}
        ,900000); // 15 minutes = 900000 ms
	getProject().getInputHandler().handleInput(request);

        t.cancel();
        String value = request.getInput();
        
	// add-on
	if (regexp != null) {
	    while (!value.matches(regexp)) {
		getProject().getInputHandler().handleInput(request);
		value = request.getInput();
	    }
	
	}
		
        if ((value == null || value.trim().length() == 0)
            && defaultvalue != null) {
            value = defaultvalue;
        }
		
        if (addproperty != null && value != null) {
            getProject().setNewProperty(addproperty, value);
        } 
		
    }

}
